/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2001-2003 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/exxon_conversion.c
*
**********************************************************************/

/*
 * A conversion routine for EXXON.
 */

#include <vdb_type.h>
#include <vdb_routines.h>
#include <vdb_errors.h>
#include <vdb_convert.h>
#include <vsys_routines.h>
#include <vhytec.h>

/*
 *********************************************************************
 * myEnv_t is a structure that we allocate and initialize at startup.
 * It holds all cparams that we need for conversion for all channels.
 *********************************************************************
 */

typedef struct {
  float min_counts;
  float max_counts;
  float min_eu;
  float max_eu;
  char* name;
} myChannel_t;

typedef struct {
  int debug;		/* Debug level. */
  myChannel_t *chan;
} myEnv_t;

/*
 *********************************************************************
 * Error handling
 *********************************************************************
 */

#define RRR() rrr (__LINE__)

static void rrr (int line) {
  printf ("fatal error at %d, %s\n", line, __FILE__);
  exit (EXIT_FAILURE);
}

/*
 *********************************************************************
 * An initialization routine example.
 *********************************************************************
 */

int init (vhytecEnv_t *env) {

  myEnv_t * my;
  unsigned i;

  /* Make sure that the versions of our modude and the vhytec scanner
     are compatible. */

  VHYTEC_CHECK_VERSION (env);

  /* Allocate our own addition to the vhytecEnv_t. */

  if ((my = calloc (1, sizeof *my)) == NULL) RRR ();
  env->user = my;

  /* Parse the userParameter (if specified). In this example it is just
     an integer number indicating a debug level. */

  if (env->commandLineParameter != NULL) {
    my->debug = atoi (env->commandLineParameter);
    if (my->debug != 0)
      printf ("userParameter = <%s>\n", env->commandLineParameter);
  }

  if (my->debug)
    printf ("Number of channels = %d\n", env->numChannels);

  /* The number of channels parameter can be used to allocate and
     prefetch information needed to perform the conversion. */

  my->chan = calloc (env->numChannels, sizeof my->chan[0]);
  if (my->chan == NULL) RRR ();

  for (i = 0; i < env->numChannels; i++) {

    myChannel_t* mc = &my->chan[i];
    vista_chix* chix = &env->channel[i]->chix;
    struct conversion_record r;
    struct vdb_descrip_v desc;
    char name[256];

    if (my->debug) {
      if (vdbc_channel_name (chix, name, sizeof name) != VDB_ERR_OK) RRR ();
      mc->name = malloc (strlen (name) + 1);
      if (mc->name == NULL) RRR ();
      strcpy (mc->name, name);
    }

    desc.dsc_a_pointer = &r;
    desc.dsc_w_length = sizeof (r);
    desc.dsc_b_class = 0;
    desc.dsc_b_dtype = 0;

    if (vdb_conversion_record (chix, &desc) != VDB_SUCCESS) RRR ();

#if 1 /* Baytown */
    mc->min_counts = r.conv_param_3;
    mc->max_counts = r.conv_param_4;
    mc->min_eu = r.conv_param_1;
    mc->max_eu = r.conv_param_2;
#else /* Baton Rouge */
    mc->min_counts = r.conv_param_1;
    mc->max_counts = r.conv_param_2;
    mc->min_eu = r.conv_param_3;
    mc->max_eu = r.conv_param_4;
#endif
  }

  return VHYTEC_OK;
}

/*
 *********************************************************************
 * An example of a conversion routine from internal hardware format
 * to external database and/or logger format.
 *********************************************************************
 */

int toExt (vhytecEnv_t *env, vhytecChannel_t *chan, vhytecValue_t *val) {

  myEnv_t *my = env->user;
  unsigned id = chan->id;
  myChannel_t *p = &my->chan[id];
  float slope;

  if (my->debug) {
    printf ("toExt routine for channel %s\n", p->name);
    switch (chan->hwDtype) {
    case VDB_DTYPE_INT16:
      printf ("Hardware value = %d\n", val->int16);
      break;
    case VDB_DTYPE_UINT16:
      printf ("Hardware value = %u\n", val->uint16);
      break;
    default:
      RRR ();
    }
  }

  slope = (p->max_eu - p->min_eu) / (p->max_counts - p->min_counts);

  switch (chan->hwDtype) {
  case VDB_DTYPE_INT16:
    val->float32 = (val->int16 - p->min_counts) * slope + p->min_eu;
    break;
  case VDB_DTYPE_UINT16:
    val->float32 = (val->uint16 - p->min_counts) * slope + p->min_eu;
    break;
  default:
    RRR ();
  }

  if (my->debug)
    printf ("Database value = %g\n", val->float32);
    
  return VHYTEC_OK;
}
