/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2000-2003 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/logtest.c
*
**********************************************************************/

/*
 * Program to test/research accuracy of logging.
 * Author: Eugene Zharkov
 * Started: 24-Nov-2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <vsys_version.h>
#include <vsys_errors.h>
#include <vsys_alloc.h>
#include <vsys_time.h>
#include <vsys_cli.h>
#include <vthread.h>
#include <vloglib.h>

static char programName[] = "Vlog Accuracy Test";
static char commandName[] = "logtest";
static char versionString[] = VSYS_VERSION_STRING "." "2";

typedef int bool;

static struct {
  bool alwaysWait;
  bool autotuneStartTime;
  bool banner;
  bool circular;
  bool fsync;
  bool restart;
  bool shared;
  bool spinSoft;
  bool useNanosleep;
  bool useSpin;
  bool useUsleep;
  vtime scanRateTime;		/* Scan rate in vtime form. */
  vtime scanRateHalfTime;	/* Scan rate in vtime form / 2 */
  char * name;			/* Log name. */
  unsigned count;		/* Repetition count. */
  unsigned bufferSize;
  unsigned initialFileSize;
  unsigned intervalSec;
  unsigned intervalKb;
  unsigned latency;
  unsigned lengthKb;
  unsigned lengthSec;
  unsigned numChannels;
  unsigned scanRate;		/* Scan rate in milliseconds. */
  unsigned workCount;		/* Repetition count for doSomething. */
  vthreadCond_t cond;
  vthreadMutex_t mutex;
  vlogApi_t * api;
  vlogLog_t * log;
  vlogChannel_t * endScanChannel;
  vlogChannel_t * startScanChannel;
  vlogChannel_t * startTimeChannel;
  vlogChannel_t ** channel;
} data;

/*
 *********************************************************************
 */

#define RRR() rrr (__LINE__)
#define FIXME() rrr (__LINE__)

static void rrr (int line) {
  printf ("%s: error at %d\n", commandName, line);
  if (data.api != NULL) vlogc_perror (data.api, NULL);
  exit (EXIT_FAILURE);
}

/*
 *********************************************************************
 */

#define timADD(t1, t2, t3) do {			\
  (t3).t_sec = (t1).t_sec + (t2).t_sec;		\
  (t3).t_nsec = (t1).t_nsec + (t2).t_nsec;	\
  if ((t3).t_nsec >= 1000000000) {		\
    (t3).t_sec += 1;				\
    (t3).t_nsec -= 1000000000;			\
  }						\
} while (0)

#define timCMP(t1, t2) (			\
  (t1).t_sec > (t2).t_sec ? 1 :			\
  (t1).t_sec < (t2).t_sec ? -1 :		\
  (t1).t_nsec > (t2).t_nsec ? 1 :		\
  (t1).t_nsec == (t2).t_nsec ? 0 : -1		\
)

#define timSUB(t1, t2, t3) (					\
  (t1).t_nsec < (t2).t_nsec ?					\
   ((t3).t_sec = (t1).t_sec - (t2).t_sec - 1,			\
    (t3).t_nsec =  (t1).t_nsec + 1000000000 - (t2).t_nsec) :	\
   ((t3).t_sec = (t1).t_sec - (t2).t_sec,			\
    (t3).t_nsec = (t1).t_nsec - (t2).t_nsec)			\
)

#define timINC(t) do {				\
  if ((t).t_nsec == 999999999) {		\
    (t).t_nsec = 0;				\
    (t).t_sec++;				\
  }						\
  else						\
    (t).t_nsec++;				\
} while (0)

/*
 *********************************************************************
 */

static void openVarchive (void) {

  vlogLogAttr_t * attr;
  vtime time;

  if ((data.api = vlogc_init ()) == NULL) RRR ();
  if ((attr = vlogc_attr_create (data.api)) == NULL) RRR ();
  if (data.restart)
    if (vlogc_attr_set_replace (attr, 1) != 1) RRR ();
  if (! data.shared && ! data.circular)
    if (vlogc_attr_set_sequential (attr, 1) != 1) RRR ();
  if (data.shared) {
    if (vlogc_attr_set_shared (attr, 1) != 1) RRR ();
    if (data.restart)
      if (vlogc_attr_set_restart (attr, 1) != 1) RRR ();
  }
  if (data.fsync)
    if (vlogc_attr_set_fsync (attr, 1) != 1) RRR ();
  if (vlogc_attr_set_buffer_size (attr, data.bufferSize) != 1) RRR ();
  if (vlogc_attr_set_file_size (attr, data.initialFileSize) != 1) RRR ();
  if ((data.log = vlogc_open (data.api, data.name, attr)) == NULL) RRR ();
  vlogc_global_lock (data.api);
  if (data.circular)
    if (vlogc_set_log_length (data.log, 1,
			      data.lengthSec, data.intervalSec,
			      data.lengthKb, data.intervalKb) != 1) RRR ();
  if (data.shared && data.latency)
    if (vlogc_set_latency (data.log, data.latency) != 1) RRR ();
  vlogc_attr_delete (attr);
  vsys_get_time (&time);

  data.startTimeChannel = vlogc_add_channel (data.log, &time, "D::STARTTIME",
						    VDB_TYPE_INT8, 1);
  if (data.startTimeChannel == NULL) RRR ();

  data.startScanChannel = vlogc_add_channel (data.log, &time, "D::STARTSCAN",
					     VDB_TYPE_INT8, 1);
  if (data.startScanChannel == NULL) RRR ();

  data.endScanChannel = vlogc_add_channel (data.log, &time, "D::ENDSCAN",
					   VDB_TYPE_TIME, 1);
  if (data.endScanChannel == NULL) RRR ();

  if (data.numChannels != 0) {
    char name[20];
    unsigned i;
    data.channel = VMEM_MALLOC (data.numChannels * sizeof data.channel[0]);
    if (data.channel == NULL) RRR ();
    for (i = 0; i < data.numChannels; i++) {
      sprintf (name, "D::Z%d", i);
      data.channel[i] = vlogc_add_channel (data.log, &time, name,
					   VDB_TYPE_INTEGER, 1);
      if (data.channel[i] == NULL) RRR ();
    }
  }
  vlogc_global_unlock (data.api);
}

/*
 *********************************************************************
 */

static void closeVarchive (void) {
  if (vlogc_close (data.log) != 1) RRR ();
  vlogc_free (data.api);
}

/*
 *********************************************************************
 */

static void logChannel (vlogChannel_t *chan, vtime *ts, void *value) {
  static int zero[2] = { 0, 0 };
  if (value == NULL) {
    value = zero;
    zero[0]++;
  }
  vlogc_global_lock (data.api);
  if (vlogc_log_channel (data.log, ts, chan, value) != 1) RRR ();
  vlogc_global_unlock (data.api);
}

/*
 *********************************************************************
 */

static void doSomething (void) {
  char s[] = { 'x', 0 };
  unsigned i;
  for (i = 0; i < data.workCount; i++) {
    vsys_strupr (s);
  }
}

/*
 *********************************************************************
 */

static void test (void) {

  struct timespec ts;
  vtime endScanTime;
  vtime nextTime;
  vtime startScanTime;
  unsigned i;

  if (vthreadMutexLock (&data.mutex)) RRR ();

  vsys_get_time (&nextTime);

  if (data.autotuneStartTime) {
    timADD (nextTime, data.scanRateTime, nextTime);
    ts.tv_nsec = nextTime.t_nsec;
    ts.tv_sec = nextTime.t_sec;
    if (vthreadCondTimedwait (&data.cond, &data.mutex, &ts) !=
	VTHREAD_ETIMEDOUT) RRR ();
    vsys_get_time (&nextTime);
    timADD (nextTime, data.scanRateTime, nextTime);
    ts.tv_nsec = nextTime.t_nsec;
    ts.tv_sec = nextTime.t_sec;
    if (vthreadCondTimedwait (&data.cond, &data.mutex, &ts) !=
	VTHREAD_ETIMEDOUT) RRR ();
    vsys_get_time (&nextTime);
  }

  logChannel (data.startTimeChannel, &nextTime, NULL);
  
  for (i = 0; i < data.count; i++) {
    vsys_get_time (&startScanTime);
    logChannel (data.startScanChannel, &startScanTime, NULL);
    doSomething ();
    if (data.numChannels != 0) {
      unsigned i;
      for (i = 0; i < data.numChannels; i++)
	logChannel (data.channel[i], &startScanTime, NULL);
    }
    timADD (nextTime, data.scanRateTime, nextTime);
    vsys_get_time (&endScanTime);
    
    if (timCMP (nextTime, endScanTime) <= 0) {
      vtime tmp;
      while (1) {
	timADD (nextTime, data.scanRateTime, tmp);
	if (timCMP (tmp, endScanTime) > 0) {
	  vtime startPlusHalf;
	  timADD (startScanTime, data.scanRateHalfTime, startPlusHalf);
	  if (timCMP (startPlusHalf, endScanTime) > 0)
	    nextTime = tmp;
	  break;
	}
	nextTime = tmp;
      }
    }
    logChannel (data.endScanChannel, &endScanTime, &nextTime);
    if (timCMP (nextTime, endScanTime) > 0) {
      vtime diffTime;
      timSUB (nextTime, endScanTime, diffTime);
      if (diffTime.t_sec == 0) {
	if (data.useSpin) {
	  vtime currentTime;
	  do {
#ifdef UNIX
	    if (data.spinSoft) sched_yield ();
#endif
	    vsys_get_time (&currentTime);
	  } while (timCMP (currentTime, nextTime) < 0);
	  continue;
	}
#ifdef UNIX
	else if (data.useUsleep) {
#if defined DUNIX || defined POWERUX
	  RRR ();
#else
	  usleep (diffTime.t_nsec / 1000);
#endif
	  continue;
	}
	else if (data.useNanosleep) {
	  ts.tv_sec = 0;
	  ts.tv_nsec = diffTime.t_nsec;
	  nanosleep (&ts, NULL);
	  continue;
	}
#endif
      }
    }
    ts.tv_nsec = nextTime.t_nsec;
    ts.tv_sec = nextTime.t_sec;
    if (vthreadCondTimedwait (&data.cond, &data.mutex, &ts) !=
	VTHREAD_ETIMEDOUT) RRR ();
  }
}

/*
 *********************************************************************
 */

int main (int argc, char *argv[]) {

  vcliPriority_t priority;
  vtime startTime, endTime, elapsedTime;
  long oper = VSYS_SECS;
  vcli_t * cli;
  double elapsed;
  int status;

  if ((cli = vcliInit ()) == NULL) RRR ();

  if (vcliSetTitle (cli, programName) != VCLI_OK ||
      vcliSetVersion (cli, versionString) != VCLI_OK ||
      vcliSetCommandName (cli, commandName) != VCLI_OK ||
      vcliEnableSortedHelp (cli) != VCLI_OK ||
      vcliVparse
      (cli,
       argc, argv,
       "define type length",
       "k seconds $uint", &data.lengthSec,
       "k KBytes $uint", &data.lengthKb,

       "define type intervalLength",
       "k seconds $uint", &data.intervalSec,
       "k KBytes $uint", &data.intervalKb,

       "q alwaysWait", &data.alwaysWait,
       "q autotuneStartTime", &data.autotuneStartTime,
       "q banner default 1", &data.banner,
       "q bufferSize $uint default 512", &data.bufferSize,
       "q circular", &data.circular,
       "q count $uint default 1000", &data.count,
       "q fileSize $uint default 512", &data.initialFileSize,
       "q fsync", &data.fsync,
       "q intervalLength type intervalLength list", NULL,
       "q latency $uint", &data.latency,
       "q length type length list", NULL,
       "q name $file default 'logtest.varc'", &data.name,
       "q numChannels $uint", &data.numChannels,
       "q priority $priority", &priority,
       "q rate $uint default 10", &data.scanRate,
       "q restart", &data.restart,
       "q shared", &data.shared,
       "q spin", &data.useSpin,
       "q spinSoft", &data.spinSoft,
       "q workCount $uint", &data.workCount,
#ifdef UNIX
       "q nanosleep", &data.useNanosleep,
       "q usleep", &data.useUsleep,
#endif
       NULL) != VCLI_OK) {
    vcliPerror (cli, NULL);
    RRR ();
  }

  if (data.banner)
    printf ("Vsystem %s Version %s\n", programName, versionString);

  if (vcliSetPriority (cli, &priority) != VCLI_OK)
    vcliPerror (cli, NULL);

  data.scanRateTime.t_sec = data.scanRate / 1000;
  data.scanRateTime.t_nsec = (data.scanRate % 1000) * 1000000;
  data.scanRateHalfTime.t_sec = data.scanRateTime.t_sec / 2;
  data.scanRateHalfTime.t_nsec = data.scanRateTime.t_nsec / 2;
  if (data.scanRateTime.t_sec & 1) data.scanRateHalfTime.t_nsec += 500000000;

  if (vthreadMutexInit (&data.mutex, NULL)) RRR ();
  if (vthreadCondInit (&data.cond, NULL)) RRR ();

  openVarchive ();
  vsys_get_time (&startTime);
  test ();
  vsys_get_time (&endTime);
  vsys_subtime (&endTime, &startTime, &elapsedTime, &status);
  vsys_conv_to_double (&oper, &elapsed, &elapsedTime, &status);
  printf ("Elapsed time = %g seconds\n", elapsed);
  vlogc_global_lock (data.api);
  closeVarchive ();
  
  return EXIT_SUCCESS;
}
