/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2000-2004 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/sim.c
*
**********************************************************************/

#include <ctype.h>
#include <vdb_type.h>
#include <vsys_internal.h>
#include "vhytec.h"
#include "internal.h"
#include "io.h"

static char errPrefix[] = "vhytec/sim";

#define SIM_N_VARS 2600
#define SIM_N_UNITS 4

typedef Vmbport vp_t;

typedef struct {
  struct {
    VUINT16 repflg;
    VUINT16 value;
  } i16, i1;
} simVar_t;

typedef struct {
  unsigned addr;
  unsigned numValues;
  void* value;
  unsigned index;
} addrValueList_t;

#define ENRON 0x1000

typedef struct table_s {
  struct table_s * next;
  int function;
  int unit;			/* unit + ENRON flag */
  addrValueList_t * list;
  unsigned listSize;
  int dtype;			/* 0 for native types or VDB_DTYPE_FLOAT,
				   or VDB_DTYPE_DOUBLE */
} table_t;

typedef struct sim1_s {
  bool doNotAutoinc;
  bool autoincAlways;
  bool bigEndian;
  unsigned i16getDelay;
  unsigned nPutsToFail;
  unsigned nGetsToFail;
  unsigned unit;
  simVar_t var[SIM_N_VARS];
  vp_t * vp;
  table_t * tableList;		/* List of tables. */
  table_t * currentTable;	/* Current table. */
} sim_t;

static int value0int[] = { 0 };

#define simFATAL() simFatal (__LINE__)
#define RRR() simFatal (__LINE__)

static void simFatal (int line) {
  printf ("%s: fatal SIM error at %d\n", errPrefix, line);
  exit (EXIT_FAILURE);
}

/*
 *********************************************************************
 */

static void setTable (vp_t *vp, table_t *tableList) {
  sim_t *sim = vp->userData;
  sim->tableList = tableList;
}

/*
 *********************************************************************
 */

static void vaxToIeeeTable (vp_t *vp) {
#ifdef VMS
  sim_t *sim = vp->userData;
  table_t *table = sim->tableList;
  while (table != NULL) {
    if (table->dtype == VDB_DTYPE_FLOAT ||
	table->dtype == VDB_DTYPE_DOUBLE) {
      addrValueList_t *list = table->list;
      unsigned i;
      for (i = 0; i < table->listSize; i++) {
	void *value = list[i].value;
	unsigned j;
	for (j = 0; j < list[i].numValues; j++) {
	  if (table->dtype == VDB_DTYPE_FLOAT) {
	    float *v = value;
	    vax_to_ieee_float (&v[j], &v[j]);
	  }
	  else {
	    double *v = value;
	    vax_to_ieee_double (&v[j], &v[j]);
	  }
	}
      }
    }
    table = table->next;
  }
#endif
}

/*
 *********************************************************************
 * Macros for defining tables
 *********************************************************************
 */

#define VALUE(type, prefix, addr) static type value##prefix##addr[]

#define ADDR_VALUE(addr, prefix) {		\
  addr,						\
  VSYS_ARRAY_SIZE (value##prefix),		\
  value##prefix					\
}

#define ADDR_PREFIX(addr, prefix) ADDR_VALUE(addr, prefix##addr)

#define M9x2s(m2,n,a2)				\
  m2 (n##1, a2);				\
  m2 (n##2, a2);				\
  m2 (n##3, a2);				\
  m2 (n##4, a2);				\
  m2 (n##5, a2);				\
  m2 (n##6, a2);				\
  m2 (n##7, a2);				\
  m2 (n##8, a2);				\
  m2 (n##9, a2)

#define M9x2c(m2,n,a2)				\
  m2 (n##1, a2),				\
  m2 (n##2, a2),				\
  m2 (n##3, a2),				\
  m2 (n##4, a2),				\
  m2 (n##5, a2),				\
  m2 (n##6, a2),				\
  m2 (n##7, a2),				\
  m2 (n##8, a2),				\
  m2 (n##9, a2)

#define M10x2s(m2,n,a2) m2 (n##0, a2); M9x2s (m2, n, a2)
#define M10x2c(m2,n,a2) m2 (n##0, a2), M9x2c (m2, n, a2)

#define M99x2c(m2,n,a2)				\
  M9x2c (m2, n##0, a2),				\
  M10x2c (m2, n##1, a2),			\
  M10x2c (m2, n##2, a2),			\
  M10x2c (m2, n##3, a2),			\
  M10x2c (m2, n##4, a2),			\
  M10x2c (m2, n##5, a2),			\
  M10x2c (m2, n##6, a2),			\
  M10x2c (m2, n##7, a2),			\
  M10x2c (m2, n##8, a2),			\
  M10x2c (m2, n##9, a2)

#define M100x2c(m2,n,a2) m2 (n##00, a2), M99x2c (m2, n, a2)

/* old macros, should no longer be used for new stuff */

#define M(a,x) { a, VSYS_ARRAY_SIZE (value##x), value##x }
#define M9s(m1,n)					\
  m1 (n##1); m1 (n##2); m1 (n##3); m1 (n##4);		\
  m1 (n##5); m1 (n##6); m1 (n##7); m1 (n##8); m1 (n##9)
#define M9c(m1,n)					\
  m1 (n##1), m1 (n##2), m1 (n##3), m1 (n##4),		\
  m1 (n##5), m1 (n##6), m1 (n##7), m1 (n##8), m1 (n##9)
#define M10s(m1,n) m1 (n##0); M9s (m1, n)
#define M10c(m1,n) m1 (n##0), M9c (m1, n)

/*
 *********************************************************************
 */

static void test10Init (vp_t *vp) {

  static double value8x29[] = { -12.23, 0}; /* binary */
  static double value8x33[] = { 2.3 }; /* int8 */
  static double value8x37[] = { 256.0 + 129.1 }; /* uint8 */
  static double value8x41[] = { -12345.2 }; /* int16 */
  static double value8x45[] = { 22222.2 }; /* uint16 */
  static double value8x49[] = { 32000.2, 65535. }; /* int32 */
  static double value8x53[] = { 31234.2, 65534.2 }; /* uint32 */
  static double value8x57[] = { 100.1 }; /* double */
  static double value8x61[] = { -34.2 }; /* float */
  static addrValueList_t list8[] = { M(28,8x29), M(32,8x33), M(36,8x37),
				     M(40,8x41), M(44,8x45), M(48,8x49),
				     M(52,8x53), M(56,8x57), M(60,8x61) };
  static table_t table8 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list8, VSYS_ARRAY_SIZE (list8), VDB_DTYPE_DOUBLE };

  static double value7x39[] = { -12.23, 0.0, 2.0, 0.0 }; /* binary */
  static double value7x43[] = { 12.3, -45.6, 258.2 }; /* int8 */
  static double value7x47[] = { 200.3, 256.0 + 129.1 }; /* uint8 */
  static double value7x51[] = { 23456.1, -12345.2 }; /* int16 */
  static double value7x55[] = { 65000.2, 22222.2 }; /* uint16 */
  static double value7x59[] = { -123.2, 32000.2, 65535. }; /* int32 */
  static double value7x63[] = { -1.1, 31234.2, 65534.2 }; /* uint32 */
  static double value7x67[] = { -999.5, 100.1 }; /* double */
  static double value7x71[] = { 23.1, -34.2 }; /* float */
  static addrValueList_t list7[] = { M(38,7x39), M(42,7x43), M(46,7x47),
				     M(50,7x51), M(54,7x55), M(58,7x59),
				     M(62,7x63), M(66,7x67), M(70,7x71) };
  static table_t table7 = { &table8, F_READ_INPUT_REGISTERS, 1,
			    list7, VSYS_ARRAY_SIZE (list7), VDB_DTYPE_DOUBLE };

  static float value6x11[] = { -12.23F, 0}; /* binary */
  static float value6x13[] = { 2.3F }; /* int8 */
  static float value6x15[] = { 256.0F + 129.1F }; /* uint8 */
  static float value6x17[] = { -12345.2F }; /* int16 */
  static float value6x19[] = { 22222.2F }; /* uint16 */
  static float value6x21[] = { 32000.2F, 65535.F }; /* int32 */
  static float value6x23[] = { 31234.2F, 65534.2F }; /* uint32 */
  static float value6x25[] = { 100.1F }; /* double */
  static float value6x27[] = { -34.2F }; /* float */
  static addrValueList_t list6[] = { M(10,6x11), M(12,6x13), M(14,6x15),
				     M(16,6x17), M(18,6x19), M(20,6x21),
				     M(22,6x23), M(24,6x25), M(26,6x27) };
  static table_t table6 = { &table7, F_READ_OUTPUT_REGISTERS, 1,
			    list6, VSYS_ARRAY_SIZE (list6), VDB_DTYPE_FLOAT };

  static float value5x21[] = { -12.23F, 0.0F, 2.0F, 0.0F }; /* binary */
  static float value5x23[] = { 12.3F, -45.6F, 258.2F }; /* int8 */
  static float value5x25[] = { 200.3F, 256.0F + 129.1F }; /* uint8 */
  static float value5x27[] = { 23456.1F, -12345.2F }; /* int16 */
  static float value5x29[] = { 65000.2F, 22222.2F }; /* uint16 */
  static float value5x31[] = { -123.2F, 32000.2F, 65535.F }; /* int32 */
  static float value5x33[] = { -1.1F, 31234.2F, 65534.2F }; /* uint32 */
  static float value5x35[] = { -999.5F, 100.1F }; /* double */
  static float value5x37[] = { 23.1F, -34.2F }; /* float */
  static addrValueList_t list5[] = { M(20,5x21), M(22,5x23), M(24,5x25),
				     M(26,5x27), M(28,5x29), M(30,5x31),
				     M(32,5x33), M(34,5x35), M(36,5x37) };
  static table_t table5 = { &table6, F_READ_INPUT_REGISTERS, 1,
			    list5, VSYS_ARRAY_SIZE (list5), VDB_DTYPE_FLOAT };

  static int value4x1[] = { 1, 0, 2, 3 }; /* binary */
  static int value4x2[] = { 12, -34, 257, 256 + 255 }; /* int8 */
  static int value4x3[] = { 23, 200, 511 }; /* uint8 */
  static int value4x4[] = { 23456, -12345 }; /* int16 */
  static int value4x5[] = { 65000, 22222, -1 }; /* uint16 */
  static int value4x6[] = { -123, 32000, 65535 }; /* int32 */
  static int value4x7[] = { -1, 31234, 65534 }; /* uint32 */
  static int value4x8[] = { -999, 100 }; /* double */
  static int value4x9[] = { 23, -34 }; /* float */
  static addrValueList_t list4[] = { M(0,4x1), M(1,4x2), M(2,4x3), M(3,4x4),
				     M(4,4x5), M(5,4x6), M(6,4x7), M(7,4x8),
				     M(8,4x9) };
  static table_t table4 = { &table5, F_READ_OUTPUT_REGISTERS, 1,
			    list4, VSYS_ARRAY_SIZE (list4) };

  static int value3[] = { 0, 1, 0, 1 };
  static int value3x[] = { 1 };
  static addrValueList_t list3[] = { M(0,3), M(1,3), M(2,3), M(3,3), M(4,3),
				     M(5,3), M(6,3), M(7,3), M(8,3),
				     M(9,3x), M(10,3x), M(11,3x), M(12,3x),
				     M(13,3x), M(14,3x), M(15,3x) };
  static table_t table3 = { &table4, F_READ_OUTPUT_DISCRETES, 1,
			    list3, VSYS_ARRAY_SIZE (list3) };

  static int value2x1[] = { 1, 0, 2, 3 }; /* binary */
  static int value2x2[] = { 12, -34, 257, 256 + 255 }; /* int8 */
  static int value2x3[] = { 23, 200, 511 }; /* uint8 */
  static int value2x4[] = { 23456, -12345 }; /* int16 */
  static int value2x5[] = { 65000, 22222, -1 }; /* uint16 */
  static int value2x6[] = { -123, 32000, 65535 }; /* int32 */
  static int value2x7[] = { -1, 31234, 65534 }; /* uint32 */
  static int value2x8[] = { -999, 100 }; /* double */
  static int value2x9[] = { 23, -34 }; /* float */
  static int value2x10[] = { 1, 0, 2, 3 }; /* binary */
  static int value2x11[] = { 12, -34, 257, 256 + 255 }; /* int8 */
  static int value2x12[] = { 23, 200, 511 }; /* uint8 */
  static int value2x13[] = { 23456, -12345 }; /* int16 */
  static int value2x14[] = { 65000, 22222, -1 }; /* uint16 */
  static int value2x15[] = { -2, 32000, 65535 }; /* int32 */
  static int value2x16[] = { -1, 31234, 65534 }; /* uint32 */
  static int value2x17[] = { -1, 60000 }; /* double */
  static int value2x18[] = { 10000, 50000 }; /* float */
  static addrValueList_t list2[] = { M(0,2x1), M(1,2x2), M(2,2x3), M(3,2x4),
				     M(4,2x5), M(5,2x6), M(6,2x7), M(7,2x8),
				     M(8,2x9), M(9,2x10), M(10,2x11),
				     M(11,2x12), M(12,2x13), M(13,2x14),
				     M(14,2x15), M(15,2x16), M(16,2x17),
				     M(17,2x18) };
  static table_t table2 = { &table3, F_READ_INPUT_REGISTERS, 1,
			    list2, VSYS_ARRAY_SIZE (list2) };

  static int value1[] = { 1, 0, 1 };
  static int value1x[] = { 1 };
  static addrValueList_t list1[] = { M(0,1), M(1,1), M(2,1), M(3,1), M(4,1),
				     M(5,1), M(6,1), M(7,1), M(8,1),
				     M(9,1x), M(10,1x), M(11,1x), M(12,1x),
				     M(13,1x), M(14,1x), M(15,1x) };
  static table_t table1 = { &table2, F_READ_INPUT_DISCRETES, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };

  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test11Init (vp_t *vp) {

#define M1(n) static int value2x##n[] = { 0 }
  M10s (M1, 10);
  M10s (M1, 11);
  M10s (M1, 12);
  M10s (M1, 13);
  M10s (M1, 14);
  M10s (M1, 15);
  M10s (M1, 16);
#undef M1
#define M1(n) M (n, 2x##n)
  static addrValueList_t list2[] = {
    M10c (M1, 10),
    M10c (M1, 11),
    M10c (M1, 12),
    M10c (M1, 13),
    M10c (M1, 14),
    M10c (M1, 15),
    M10c (M1, 16)
  };
#undef M1
  static table_t table2 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list2, VSYS_ARRAY_SIZE (list2) };

  static int value1x113[] = { 0 };
  static int value1x114[] = { 0 };
  static int value1x115[] = { 0 };
  static int value1x116[] = { 0 };
  static int value1x117[] = { 0 };
  static int value1x118[] = { 0 };
  static int value1x119[] = { 0 };
  static int value1x120[] = { 0 };
  static int value1x121[] = { 0 };
  static int value1x122[] = { 0 };
  static int value1x123[] = { 0 };
  static int value1x124[] = { 0 };
  static int value1x125[] = { 0 };
  static int value1x126[] = { 0 };
  static int value1x127[] = { 0 };
  static int value1x128[] = { 0 };
  static int value1x129[] = { 0 };
  static int value1x130[] = { 0 };
  static int value1x131[] = { 0 };
  static int value1x132[] = { 0 };
  static int value1x133[] = { 0 };
  static int value1x134[] = { 0 };
  static int value1x135[] = { 0 };
  static int value1x136[] = { 0 };
  static int value1x137[] = { 0 };
  static int value1x138[] = { 0 };
  static int value1x139[] = { 0 };
  static int value1x140[] = { 0 };
  static int value1x141[] = { 0 };
  static int value1x142[] = { 0 };
  static int value1x143[] = { 0 };
  static int value1x144[] = { 0 };
  static int value1x145[] = { 0 };
  static addrValueList_t list1[] = {
    M (112, 1x113),
    M (113, 1x114),
    M (114, 1x115),
    M (115, 1x116),
    M (116, 1x117),
    M (117, 1x118),
    M (118, 1x119),
    M (119, 1x120),
    M (120, 1x121),
    M (121, 1x122),
    M (122, 1x123),
    M (123, 1x124),
    M (124, 1x125),
    M (125, 1x126),
    M (126, 1x127),
    M (127, 1x128),
    M (128, 1x129),
    M (129, 1x130),
    M (130, 1x131),
    M (131, 1x132),
    M (132, 1x133),
    M (133, 1x134),
    M (134, 1x135),
    M (135, 1x136),
    M (136, 1x137),
    M (137, 1x138),
    M (138, 1x139),
    M (139, 1x140),
    M (140, 1x141),
    M (141, 1x142),
    M (142, 1x143),
    M (143, 1x144),
    M (144, 1x145)
  };
  static table_t table1 = { &table2, F_READ_OUTPUT_DISCRETES, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };

  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test12Init (vp_t *vp) {
  static int value1x2[] = { 0 };
  static addrValueList_t list1[] = { M (1, 1x2) };
  static table_t table1 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };
  setTable (vp, &table1);
}

/*
 *********************************************************************
 * This is also used in test22.vva & test25.vva
 *********************************************************************
 */

static void test15Init (vp_t *vp) {
  static int value2x1[] = { 1, 0, 1, 1 };
  static int value2x2[] = { 0, 1, 0, 1 };
  static int value2x3[] = { 1, 0, 1, 0 };
  static addrValueList_t list2[] = {
    M (0, 2x1),
    M (1, 2x2),
    M (2, 2x3),
  };
  static table_t table2 = { NULL, F_READ_INPUT_DISCRETES, 1,
			    list2, VSYS_ARRAY_SIZE (list2) };

  static int value1x1[] = { 11, 12, 13 };
  static int value1x2[] = { 21, 22, 23 };
  static int value1x3[] = { 31, 32, 33 };
  static addrValueList_t list1[] = {
    M (0, 1x1),
    M (1, 1x2),
    M (2, 1x3),
  };
  static table_t table1 = { &table2, F_READ_INPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };
  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test16Init (vp_t *vp) {

  /* float64 */
  static double value4x400[] = { 3.3, 4.4 };
  static addrValueList_t list4[] = { M (400, 4x400) };
  static table_t table4 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list4, VSYS_ARRAY_SIZE (list4), VDB_DTYPE_DOUBLE };
  /* float32 */
  static float value3x300[] = { 1.1F, 2.2F };
  static addrValueList_t list3[] = { M (300, 3x300) };
  static table_t table3 = { &table4, F_READ_OUTPUT_REGISTERS, 1,
			    list3, VSYS_ARRAY_SIZE (list3), VDB_DTYPE_FLOAT };
  /* binary */
  static int value2x200[] = { 0x0001, 0x0002, 0x0003 };
  static addrValueList_t list2[] = { M (200, 2x200) };
  static table_t table2 = { &table3, F_READ_OUTPUT_REGISTERS, 1,
			    list2, VSYS_ARRAY_SIZE (list2) };
  /* int16 */
  static int value1x100[] = { 11 };
  static addrValueList_t list1[] = { M (100, 1x100) };
  static table_t table1 = { &table2, F_READ_OUTPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };
  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test20Init (vp_t *vp) {

  static int value1x0[] = { 0x1401, 0x2802, 0x3c05 };
  static int value0[] = { 0x0000 };
  static int value1x130[] = { 0x1401, 0x3c03 };

  static addrValueList_t list1[] = {
    M (0, 1x0),
    M (1, 0),
    M (2, 0),
    M (3, 0),
    M (4, 0),
    M (5, 0),
    M (6, 0),
    M (7, 0),
    M (8, 0),
    M (9, 0),
#define M1(n) M (n, 0)
    M10c (M1, 1),
    M10c (M1, 2),
    M10c (M1, 3),
    M10c (M1, 4),
    M10c (M1, 5),
    M10c (M1, 6),
    M10c (M1, 7),
    M10c (M1, 8),
    M10c (M1, 9),
    M10c (M1, 10),
    M10c (M1, 11),
    M10c (M1, 12),
    M (130, 1x130),
#undef M1
  };
  static table_t table1 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };
  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test21Init (vp_t *vp) {

#define M1(n) static double value##n[] = { n##.1, n##.2 }
  M1 (0);
  M1 (1);
  M1 (2);
  M1 (3);
  M1 (4);
  M1 (5);
  M1 (6);
  M1 (7);
  M1 (8);
  M1 (9);
  M10s (M1, 1);
  M10s (M1, 2);
  M10s (M1, 3);
  M10s (M1, 4);
  M10s (M1, 5);
  M10s (M1, 6);
  M10s (M1, 7);
  M10s (M1, 8);
  M10s (M1, 9);
  M10s (M1, 10);
  M10s (M1, 11);
  M10s (M1, 12);
  M1 (130);
  M1 (131);
  M1 (132);
  M1 (133);
  M1 (134);
  M1 (135);
  M1 (136);
#undef M1

  static addrValueList_t list1[] = {
#define M1(n) M (n*4, n)
    M1 (0),
    M1 (1),
    M1 (2),
    M1 (3),
    M1 (4),
    M1 (5),
    M1 (6),
    M1 (7),
    M1 (8),
    M1 (9),
    M10c (M1, 1),
    M10c (M1, 2),
    M10c (M1, 3),
    M10c (M1, 4),
    M10c (M1, 5),
    M10c (M1, 6),
    M10c (M1, 7),
    M10c (M1, 8),
    M10c (M1, 9),
    M10c (M1, 10),
    M10c (M1, 11),
    M10c (M1, 12),
    M1 (130),
    M1 (131),
    M1 (132),
    M1 (133),
    M1 (134),
    M1 (135),
    M1 (136)
#undef M1
  };
  static table_t table1 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1), VDB_DTYPE_DOUBLE };
  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

static void test24Init (vp_t *vp) {

  static double value8x9046[] = { 80.0, 81.0 };
  static double value8x9050[] = { 82.0, 83.0 };

  static addrValueList_t list8[] = { M (9045, 8x9046), M (9049, 8x9050) };
  static table_t table8 = { NULL, F_READ_OUTPUT_REGISTERS, 1,
			    list8, VSYS_ARRAY_SIZE (list8), VDB_DTYPE_DOUBLE };

  static float value7x8048[] = { 70.0F, 71.0F };
  static float value7x8050[] = { 72.0F, 73.0F };

  static addrValueList_t list7[] = {
    M (8047, 7x8048),
    M (8049, 7x8050)
  };
  static table_t table7 = { &table8, F_READ_OUTPUT_REGISTERS, 1,
			    list7, VSYS_ARRAY_SIZE (list7), VDB_DTYPE_FLOAT };

  static double value6x6046[] = { 60.0, 61.0 };
  static double value6x6050[] = { 62.0, 63.0 };

  static addrValueList_t list6[] = { M (6045, 6x6046), M (6049, 6x6050) };
  static table_t table6 = { &table7, F_READ_OUTPUT_REGISTERS, 1,
			    list6, VSYS_ARRAY_SIZE (list6), VDB_DTYPE_DOUBLE };

  static float value5x5046[] = { 50.0F, 51.0F };
  static float value5x5048[] = { 0.0F, 0.0F };
  static float value5x5050[] = { 52.0F, 53.0F };

  static addrValueList_t list5[] = {
    M (5045, 5x5046),
    M (5047, 5x5048),
    M (5049, 5x5050)
  };
  static table_t table5 = { &table6, F_READ_OUTPUT_REGISTERS, 1,
			    list5, VSYS_ARRAY_SIZE (list5), VDB_DTYPE_FLOAT };

  VALUE (int, 4x, 10045) = { 46, 146 };
  VALUE (int, 4x, 10046) = { 0, 0 };
  VALUE (int, 4x, 10047) = { 48, 148 };

  static addrValueList_t list4[] = {
    ADDR_PREFIX (10045, 4x),
    ADDR_PREFIX (10046, 4x),
    ADDR_PREFIX (10047, 4x)
  };
  static table_t table4 = { &table5, F_READ_OUTPUT_REGISTERS, 1,
			    list4, VSYS_ARRAY_SIZE (list4) };

  static int value3x2000[] = { 11, 22 };
  static int value3x2001[] = { 0, 0 };
  static int value3x2002[] = { 111, 222 };
  static int value3x3000[] = { 3, 33 };
  static int value3x3211[] = { 44 };
  static int value3x3283[] = { 84, 184 };
  static int value3x3284[] = { 85, 185 };
  static int value3x3285[] = { 86, 186 };
  static int value3x3286[] = { 87, 187 };
  static int value3x4344[] = { 45, 145 };

  static addrValueList_t list3[] = {
    ADDR_PREFIX (2000, 3x),
    ADDR_PREFIX (2001, 3x),
    ADDR_PREFIX (2002, 3x),
    ADDR_PREFIX (3000, 3x),
    M9x2c (ADDR_VALUE, 300, 0int),
    M99x2c (ADDR_VALUE, 30, 0int),
    M100x2c (ADDR_VALUE, 31, 0int),
    M10x2c (ADDR_VALUE, 320, 0int),
    ADDR_VALUE (3210, 0int),
    ADDR_PREFIX (3211, 3x),
    ADDR_PREFIX (3283, 3x),
    ADDR_PREFIX (3284, 3x),
    ADDR_PREFIX (3285, 3x),
    ADDR_PREFIX (3286, 3x),
    ADDR_PREFIX (4344, 3x)
  };
  static table_t table3 = { &table4, F_READ_OUTPUT_REGISTERS, 1,
			    list3, VSYS_ARRAY_SIZE (list3) };

  static int value2x2046[] = { 1, 2 };

  static addrValueList_t list2[] = { M (2045, 2x2046) };
  static table_t table2 = { &table3, F_READ_OUTPUT_REGISTERS, 1,
			    list2, VSYS_ARRAY_SIZE (list2) };

  static int value1x1046[] = { 5, 2 };
  static int value1x1047[] = { 7, 0x77 };
  static int value1x1048[] = { 8, 0x88 };
  static int value1x1049[] = { 9, 0x99 };
  static int value1x1050[] = { 0x8000, 0x0, 0x8000 };
  static int value1x1066[] = { 2, 1 };

  static addrValueList_t list1[] = {
    M (1045, 1x1046),
    M (1046, 1x1047),
    M (1047, 1x1048),
    M (1048, 1x1049),
    M (1049, 1x1050),
    M (1065, 1x1066)
  };
  static table_t table1 = { &table2, F_READ_OUTPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1) };

  setTable (vp, &table1);
}

/*
 *********************************************************************
 * enron and endian
 *********************************************************************
 */

static void test27Init (vp_t *vp) {

  /* 2.4 = 4019999a */
  /* 10.8 = 412ccccd */
  /* 100.44 = 42c8e148 */

  /* enron, big endian */

  static int value5x7002[] = { 0x999a };
  static int value5x7001[] = { 0x4019 };
  static int value5x7004[] = { 0xcccd };
  static int value5x7003[] = { 0x412c };
  static int value5x7005[] = { 0 };
  static int value5x7006[] = { 0 };
  static int value5x7007[] = { 0 };
  static int value5x7008[] = { 0 };
  static int value5x7010[] = { 0xe148 };
  static int value5x7009[] = { 0x42c8 };

  static addrValueList_t list5[] = {
    M (7001, 5x7001),
    M (7002, 5x7002),
    M (7003, 5x7003),
    M (7004, 5x7004),
    M (7005, 5x7005),
    M (7006, 5x7006),
    M (7007, 5x7007),
    M (7008, 5x7008),
    M (7009, 5x7009),
    M (7010, 5x7010)
  };
  static table_t table5 = { NULL, F_READ_INPUT_REGISTERS, ENRON + 5,
			    list5, VSYS_ARRAY_SIZE (list5), 0 };

  /* enron, big endian */

  static int value4x7002[] = { 0x999a };
  static int value4x7001[] = { 0x4019 };
  static int value4x7004[] = { 0xcccd };
  static int value4x7003[] = { 0x412c };
  static int value4x7010[] = { 0xe148 };
  static int value4x7009[] = { 0x42c8 };

  static addrValueList_t list4[] = {
    M (7001, 4x7001),
    M (7002, 4x7002),
    M (7003, 4x7003),
    M (7004, 4x7004),
    M (7009, 4x7009),
    M (7010, 4x7010)
  };
  static table_t table4 = { &table5, F_READ_INPUT_REGISTERS, ENRON + 4,
			    list4, VSYS_ARRAY_SIZE (list4), 0 };

  /* enron, little endian */

  static int value3x7001[] = { 0x999a };
  static int value3x7002[] = { 0x4019 };
  static int value3x7003[] = { 0xcccd };
  static int value3x7004[] = { 0x412c };
  static int value3x7009[] = { 0xe148 };
  static int value3x7010[] = { 0x42c8 };

  static addrValueList_t list3[] = {
    M (7001, 3x7001),
    M (7002, 3x7002),
    M (7003, 3x7003),
    M (7004, 3x7004),
    M (7009, 3x7009),
    M (7010, 3x7010)
  };
  static table_t table3 = { &table4, F_READ_INPUT_REGISTERS, ENRON + 3,
			    list3, VSYS_ARRAY_SIZE (list3), 0 };

  /* standard, big endian */

  static int value2x7002[] = { 0x999a };
  static int value2x7001[] = { 0x4019 };
  static int value2x7004[] = { 0xcccd };
  static int value2x7003[] = { 0x412c };
  static int value2x7010[] = { 0xe148 };
  static int value2x7009[] = { 0x42c8 };

  static addrValueList_t list2[] = {
    M (7001, 2x7001),
    M (7002, 2x7002),
    M (7003, 2x7003),
    M (7004, 2x7004),
    M (7009, 2x7009),
    M (7010, 2x7010)
  };
  static table_t table2 = { &table3, F_READ_INPUT_REGISTERS, 2,
			    list2, VSYS_ARRAY_SIZE (list2), 0 };

  /* standard, little endian */

  static int value1x7001[] = { 0x999a };
  static int value1x7002[] = { 0x4019 };
  static int value1x7003[] = { 0xcccd };
  static int value1x7004[] = { 0x412c };
  static int value1x7009[] = { 0xe148 };
  static int value1x7010[] = { 0x42c8 };

  static addrValueList_t list1[] = {
    M (7001, 1x7001),
    M (7002, 1x7002),
    M (7003, 1x7003),
    M (7004, 1x7004),
    M (7009, 1x7009),
    M (7010, 1x7010)
  };
  static table_t table1 = { &table2, F_READ_INPUT_REGISTERS, 1,
			    list1, VSYS_ARRAY_SIZE (list1), 0 };

  setTable (vp, &table1);
}

/*
 *********************************************************************
 */

#undef M

/*
 *********************************************************************
 */

static void simInit (vp_t *vp) {

  sim_t *sim;
  short s = 1;

  vp->userData = VMEM_CALLOC (sizeof (sim_t) * SIM_N_UNITS, 1);
  if (vp->userData == NULL) simFATAL ();
  sim = vp->userData;
  sim->bigEndian = *(char*) &s == 0;


  if (isdigit (vp->name[0])) {
    switch (atoi (vp->name)) {
    case 10: test10Init (vp); break;
    case 11: test11Init (vp); break;
    case 12: test12Init (vp); break;
    case 15: test15Init (vp); break;
    case 16: test16Init (vp); break;
    case 20: test20Init (vp); break;
    case 21: test21Init (vp); break;
    case 24: test24Init (vp); break;
    case 27: test27Init (vp); break;
    default: simFATAL ();
    }
    vaxToIeeeTable (vp);
  }
  else {
    unsigned i, u;
    for (u = 0; u < SIM_N_UNITS; u++) {
      sim_t *sim = vp->userData;
      sim += u;
      sim->vp = vp;
      sim->unit = u;
      for (i = 0; i < SIM_N_VARS; i++) {
	sim->var[i].i16.value = i;
	sim->var[i].i1.value = i;
      }
    }
  }
}

/*
 *********************************************************************
 */

static addrValueList_t * findAddrValueList (sim_t *sim, table_t *table,
					    unsigned addr) {

  int function = table->function;
  int slave = table->unit & 0xff;

  while (table != NULL) {

    unsigned i;

    sim->currentTable = table;

    for (i = 0; i < table->listSize; i++) {
      addrValueList_t *list = &table->list[i];
      if (list->addr == addr)
	return list;
      if (table->dtype == VDB_DTYPE_FLOAT && addr == list->addr + 1)
	return list;
      if (table->dtype == VDB_DTYPE_DOUBLE && (addr == list->addr + 1 ||
					       addr == list->addr + 2 || 
					       addr == list->addr + 3))
	return list;
    }

    for (table = table->next; table != NULL; table = table->next) {
      if ((table->unit & 0xff) == slave &&
	  table->function == function)
	break;
    }
  }

  printf ("addr = %d, function = %d\n", addr, function);
  simFATAL ();
  return NULL;
}

/*
 *********************************************************************
 */

static int simGetFailed (sim_t *sim) {
  sim->nGetsToFail--;
  vmbSet (sim->vp, VMB_O_ERROR, "RCVTMO");
  return -1;
}

static int simPutFailed (sim_t *sim) {
  sim->nPutsToFail--;
  vmbSet (sim->vp, VMB_O_ERROR, "BADCRC");
  return -1;
}

static simVar_t * simFind (sim_t *sim, unsigned addr) {
  if (addr >= SIM_N_VARS) simFATAL ();
  return &sim->var[addr];
}

/*
 *********************************************************************
 */

static int simI16get (sim_t *sim, unsigned addr0, VUINT16 *value) {

  if (sim->tableList != NULL) {
    table_t *table = sim->currentTable;
    addrValueList_t *list = findAddrValueList (sim, table, addr0);
    table = sim->currentTable;
    if (table->dtype == VDB_DTYPE_FLOAT) {
      unsigned index = (list->index < list->numValues ?
			list->index : list->numValues - 1);
      float *v = list->value;
      union { float f; VUINT16 s[2]; } u;
      unsigned i16 = addr0 - list->addr;
      if (sim->bigEndian)
	i16 ^= 1;
      u.f = v[index];
      *value = u.s[i16];
      if (addr0 - list->addr == 1) {
	if (list->index < list->numValues)
	  list->index++;
      }
    }
    else if (table->dtype == VDB_DTYPE_DOUBLE) {
      unsigned index = (list->index < list->numValues ?
			list->index : list->numValues - 1);
      double *v = list->value;
      union { double d; VUINT16 s[4]; } u;
      unsigned i16 = addr0 - list->addr;
      if (sim->bigEndian)
	i16 ^= 3;
      u.d = v[index];
      *value = u.s[i16];
      if (addr0 - list->addr == 3) {
	if (list->index < list->numValues)
	  list->index++;
      }
    }
    else {
      unsigned index = (list->index < list->numValues ?
			list->index++ : list->numValues - 1);
      unsigned *v = list->value;
      *value = v[index];
    }
  }
  else {
    unsigned addr = addr0 + 1;
    simVar_t * var = simFind (sim, addr);
    if (sim->nGetsToFail) return simGetFailed (sim);
    if (sim->i16getDelay) {
      vsys_wait (sim->i16getDelay * 1000);
      sim->i16getDelay = 0;
    }
    *value = var->i16.value;
    if (! sim->doNotAutoinc) {
      var->i16.repflg++;
      if (sim->autoincAlways || (var->i16.repflg & 7U) != (addr & 7))
	var->i16.value += 1;
    }
    if (sim->unit != 1) *value += sim->unit * 100;
  }

  return 0;
}

/*
 *********************************************************************
 */

static int simBget (sim_t *sim, unsigned addr0, unsigned *value) {

  if (sim->tableList != NULL) {
    table_t *table = sim->currentTable;
    addrValueList_t *list = findAddrValueList (sim, table, addr0);
    unsigned index = (list->index < list->numValues ?
		      list->index++ : list->numValues - 1);
    unsigned *v = list->value;
    *value = v[index];
  }
  else {
    unsigned addr = addr0 + 1;
    simVar_t * var = simFind (sim, addr);
    if (sim->nGetsToFail) return simGetFailed (sim);
    *value = var->i1.value & 1;
    if (! sim->doNotAutoinc) {
      var->i1.repflg++;
      if (sim->autoincAlways || (var->i1.repflg & 7U) != (addr & 7))
	var->i1.value += 1;
    }
  }

  return 0;
}

/*
 *********************************************************************
 */

static int simBput (sim_t *sim, unsigned addr0, unsigned value) {

  if (sim->tableList != NULL) {
    table_t *table = sim->currentTable;
    addrValueList_t *list = findAddrValueList (sim, table, addr0);
    unsigned index = list->numValues - 1;
    unsigned *v = list->value;
    v[index] = value;
  }
  else {
    unsigned addr = addr0 + 1;
    simVar_t * var = simFind (sim, addr);
    if (sim->nPutsToFail) return simPutFailed (sim);
    var->i16.value += value ? 20 : 10;
  }

  return 0;
}

/*
 *********************************************************************
 */

static int simI16put (sim_t *sim, unsigned addr, VUINT16 value) {

  if (sim->tableList != NULL) {
    table_t *table = sim->currentTable;
    addrValueList_t *list = findAddrValueList (sim, table, addr);
    unsigned index = list->numValues - 1;
    unsigned *v = list->value;
    v[index] = value;
  }
  else {
    simVar_t * var;

    addr++;
    if (addr >= SIM_N_VARS) {
      if (addr == 3000) {
	switch (value) {
	case 1:
	  sim->doNotAutoinc = 1;
	  break;
	case 2:
	  sim->autoincAlways = 1;
	  break;
	default:
	  simFATAL ();
	}
      }
      else if (addr == 8000) {
	sim->i16getDelay = value;
      }
      else if (addr == 8001) {
	sim->nPutsToFail = value;
      }
      else if (addr == 8002) {
	sim->nGetsToFail = value;
      }
      else if (addr > 5000) {
	addr -= 5000;
	if (sim->nPutsToFail) return simPutFailed (sim);
	var = simFind (sim, addr);
	var->i1.value = value;
      }
      else if (addr > 3000) {
	addr -= 3000;
	if (sim->nPutsToFail) return simPutFailed (sim);
	var = simFind (sim, addr);
	var->i16.value = value;
      }
    }
    else {
      var = simFind (sim, addr);
      if (sim->nPutsToFail) return simPutFailed (sim);
      var->i16.value += value * sim->unit;
    }
  }

  return 0;
}

/*
 *********************************************************************
 */

static int simTransact (vp_t *vp, int slave, unsigned requestLen,
			unsigned expectedResponseLen) {

  unsigned i, addr, count, byteCount, loopCount;
  VUINT8 *response = vmbGetPtr (vp, VMB_P_USER_RESPONSE_BUFFER);
  VUINT8 *request = vmbGetPtr (vp, VMB_P_USER_REQUEST_BUFFER);
  sim_t *sim = vp->userData;
  int function = request[0];
  bool enron27;	/* enron mode in test 27 */

  if (sim->tableList != NULL) {

    table_t *table = sim->tableList;
    int tableFunction;

    switch (function) {
    case F_WRITE_OUTPUT_REGISTERS:
    case F_WRITE_OUTPUT_REGISTER:
      tableFunction = F_READ_OUTPUT_REGISTERS;
      break;
    case F_WRITE_OUTPUT_DISCRETE:
      tableFunction = F_READ_OUTPUT_DISCRETES;
      break;
    default:
      tableFunction = function;
    }

    while (table != NULL) {
      if ((table->unit & 0xff) == slave &&
	  table->function == tableFunction)
	break;
      table = table->next;
    }

    if (table == NULL) {
      printf ("function = %d, slave = %d\n", tableFunction, slave);
      simFATAL ();
    }
    sim->currentTable = table;
  }
  else {
    if (slave >= SIM_N_UNITS) simFATAL ();
    sim += slave;
  }

  enron27 = (sim->currentTable != NULL &&
	     (sim->currentTable->unit & ENRON) != 0);
  response[0] = request[0];

  switch (function) {
  case F_READ_INPUT_DISCRETES:
  case F_READ_OUTPUT_DISCRETES:
    addr = G16 (request + 1);
    count = G16 (request + 3);
    byteCount = (count + 7) / 8;
    if (addr & 15) simFATAL ();
    response[1] = byteCount;
    response += 2;
    memset (response, 0, byteCount);
    loopCount = sim->tableList == NULL ? byteCount * 8 : count;
    for (i = 0; i < loopCount ; i++) {
      unsigned value;
      if (i == 0 && sim->tableList == NULL)
	if (simBget (sim, SIM_N_VARS - 2, &value)) return -1;
      if (simBget (sim, addr, &value)) return -1;
      if (value) P1 (response, i);
      addr++;
    }
    break;
  case F_READ_INPUT_REGISTERS:
  case F_READ_OUTPUT_REGISTERS:
    addr = G16 (request + 1);
    count = G16 (request + 3);
    if (enron27) {
      count *= 2;
      if (addr < 7001) RRR ();
      addr = 7001 + (addr - 7001) * 2;
    }
    response[1] = count * 2;
    response += 2;
    for (i = 0; i < count; i++) {
      VUINT16 value;
      if (simI16get (sim, addr, &value)) return -1;
      P16 (response, value);
      response += 2;
      addr++;
    }
    break;
  case F_WRITE_OUTPUT_DISCRETE: {
    unsigned value = request[3];
    addr = G16 (request + 1);
    if (simBput (sim, addr, value)) return -1;
    memcpy (response, request, 5);
    break;
  }
  case F_WRITE_OUTPUT_REGISTER: {
    VUINT16 value = G16 (request + 3);
    addr = G16 (request + 1);
    if (simI16put (sim, addr, value)) return -1;
    memcpy (response, request, 5);
    break;
  }
  case F_WRITE_OUTPUT_REGISTERS:
    addr = G16 (request + 1);
    count = G16 (request + 3);
    memcpy (response, request, 5);
    request += 6;
    for (i = 0; i < count; i++) {
      VUINT16 value = G16 (request);
      if (simI16put (sim, addr, value)) return -1;
      request += 2;
      addr++;
    }
    break;
  default:
    printf ("function = %d\n", request[0]);
    simFATAL ();
  }

#undef P1
#undef G16
#undef P16

  return 0;
}

/*
 *********************************************************************
 */

int userConnect (vp_t *vp) {
  simInit (vp);
  return 0;
}

/*
 *********************************************************************
 */

int userTransact (vp_t *vp, int slave, unsigned requestLen,
		  unsigned expectedResponseLen) {
  return simTransact (vp, slave, requestLen, expectedResponseLen);
}

/*
 *********************************************************************
 * Conversion routines for test #12
 *********************************************************************
 */

static int user12 = 123;

int init12 (vhytecEnv_t *env) {
  VHYTEC_CHECK_VERSION (env);
  if (env->commandLineParameter == NULL) RRR ();
  if (strcmp (env->commandLineParameter, "uP") != 0) RRR ();
  if (env->numChannels != 2) RRR ();
  env->user = &user12;
  return VHYTEC_OK;
}

int toExt12 (vhytecEnv_t *env, vhytecChannel_t *chan, vhytecValue_t *val) {
  if (env->user != (void*) &user12) RRR ();
  if (chan->chix.chix != 1) RRR ();
  if (chan->id != 0) RRR ();
  if (chan->hwDtype != VDB_DTYPE_INT16) RRR ();
  if (chan->dbDtype != VDB_DTYPE_INT) RRR ();
  val->int32 = val->int16 * 3;
  return VHYTEC_OK;
}

int toInt12 (vhytecEnv_t *env, vhytecChannel_t *chan, vhytecValue_t *val) {
  union { int i; short s[2]; } u;
  if (chan->chix.chix != 2) RRR ();
  if (chan->id != 1) RRR ();
  if (chan->hwDtype != VDB_DTYPE_INT16) RRR ();
  if (chan->dbDtype != VDB_DTYPE_INT) RRR ();
  u.i = val->int32;
  val->int16 = u.s[0] * u.s[1];
  return VHYTEC_OK;
}
