/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2000-2002 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/socketiotest.c
*
**********************************************************************/

/*
 * A socket I/O test program
 * Author: Eugene Zharkov
 * Started: 29-Nov-2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <vsys_version.h>
#include <vsys_errors.h>
#include <vsys_alloc.h>
#include <vsys_time.h>
#include <vsys_cli.h>
#include <vsocket.h>

#include "version.h"

static char *programName = "Socket I/O Test";
static char commandName[] = "socketiotest";
static char versionString[] = VSYS_VERSION_STRING ".1";

typedef int bool;

static struct {
  int port;		/* TCP port number. */
  bool once;		/* Server exit after one connection. */
  bool isServer;
  bool tcpNoDelay;
  char * buffer;
  char * serverNode;
  unsigned blockSize;
  unsigned count;
  vsocklib_t * vsr;
} data;

/*
 *********************************************************************
 */

static void tsPrint (char *format, ...) {
  va_list args;
  char asctim[50];
  vtime bintim;
  vsys_get_time (&bintim);
  asctim[0] = 0;
  vsysc_asctime (asctim, sizeof (asctim), &bintim);
  printf ("%s: ", asctim);
  va_start (args, format);
  vprintf (format, args);
  printf ("\n");
  va_end (args);
}

/*
 *********************************************************************
 */

#define RRR() rrr (__LINE__)
#define FIXME() rrr (__LINE__)

static void rrr (int line) {
  tsPrint ("vhytec/server: error at %d", line);
  exit (EXIT_FAILURE);
}

/*
 *********************************************************************
 */

static double getTimeDiff (vtime *t1, vtime *t2) {

  long oper = VSYS_SECS;
  vtime diffTime;
  double diff;
  int status;

  vsys_subtime (t1, t2, &diffTime, &status);
  vsys_conv_to_double (&oper, &diff, &diffTime, &status);
  return diff;
}

/*
 *********************************************************************
 */

static void client (void) {

  vsocket_t *socket;
  vtime startTime, endTime;
  double totalBytes;
  double timeDiff;
  unsigned i;

  if (vsocketOpenClientSocket (data.vsr, data.serverNode, data.port, &socket)
    != VSOCKET_OK) RRR ();
  if (data.tcpNoDelay) {
    tsPrint ("Setting TCP_NODELAY");
    if (vsocketEnableTcpNoDelay (socket) != VSOCKET_OK) RRR ();
  }

  vsys_get_time (&startTime);

  for (i = 0; i < data.count; i++)
    if (vsocketWriteAll (socket, data.buffer, data.blockSize) != VSOCKET_OK)
      RRR ();

  if (vsocketClose (socket) != VSOCKET_OK) RRR ();
  
  vsys_get_time (&endTime);
  timeDiff = getTimeDiff (&endTime, &startTime);
  totalBytes = (double) data.count;
  totalBytes *= data.blockSize;

  tsPrint ("Block size: %u", data.blockSize);
  tsPrint ("Number of blocks sent: %u", data.count);
  tsPrint ("Number of bytes: %g", totalBytes);
  tsPrint ("MBytes/second: %g", totalBytes / timeDiff / 1000000.);
  tsPrint ("MBits/second: %g", totalBytes / timeDiff / 1000000. * 8.);
}

/*
 *********************************************************************
 */

static void runClientSession (vsocket_t *socket) {

  vtime startTime, endTime;
  unsigned count = 0;
  double totalBytes;
  double timeDiff;

  vsys_get_time (&startTime);

  while (1) {
    if (vsocketReadAll (socket, data.buffer, data.blockSize) != VSOCKET_OK) {
      if (vsocketIsEof (socket)) break;
      RRR ();
    }
    if (data.buffer[0] != 1) RRR ();
    if (data.buffer[data.blockSize - 1] != (char) data.blockSize) RRR ();
    count++;
  }

  vsys_get_time (&endTime);
  timeDiff = getTimeDiff (&endTime, &startTime);
  totalBytes = (double) count;
  totalBytes *= data.blockSize;

  tsPrint ("Block size: %u", data.blockSize);
  tsPrint ("Number of blocks received: %u", count);
  tsPrint ("Number of bytes: %g", totalBytes);
  tsPrint ("MBytes/second: %g", totalBytes / timeDiff / 1000000.);
  tsPrint ("MBits/second: %g", totalBytes / timeDiff / 1000000. * 8.);
}

/*
 *********************************************************************
 */

static void server (void) {

  vsocket_t *socket;
  vsocket_t *clientSocket;

  if (vsocketOpenServerSocket (data.vsr, data.port, &socket)
      != VSOCKET_OK) RRR ();

  if (data.tcpNoDelay) {
    tsPrint ("Setting TCP_NODELAY");
    if (vsocketEnableTcpNoDelay (socket) != VSOCKET_OK) RRR ();
  }

  while (1) {
    tsPrint ("Accepting connection");
    if (vsocketAccept (socket, &clientSocket) != VSOCKET_OK) RRR ();
    if (data.tcpNoDelay) {
      tsPrint ("Setting TCP_NODELAY");
      if (vsocketEnableTcpNoDelay (clientSocket) != VSOCKET_OK) RRR ();
    }
    tsPrint ("Accepted connection");
    runClientSession (clientSocket);
    if (vsocketClose (clientSocket) != VSOCKET_OK) RRR ();
    tsPrint ("Closed connection");
    if (data.once) {
      tsPrint ("Exiting");
      exit (EXIT_SUCCESS);
    }
  }
}

/*
 *********************************************************************
 */

static void parseCommandLine (int argc, char *argv[]) {

  vcli_t * cli;

  if ((cli = vcliInit ()) == NULL) RRR ();

  if (vcliSetTitle (cli, programName) != VCLI_OK ||
      vcliSetVersion (cli, versionString) != VCLI_OK ||
      vcliSetCommandName (cli, commandName) != VCLI_OK ||
      vcliEnableSortedHelp (cli) != VCLI_OK ||
      vcliVparse
      (cli,
       argc, argv,
       "define type programType",
       "k server subsyntax server", (void*) 1,
       "k client subsyntax client", (void*) 0,

       "q port $int default 9999", &data.port,
       "q tcpNoDelay", &data.tcpNoDelay,
       "p programType type programType", &data.isServer,
       "q blockSize $uint range 1:1000000000 default 512", &data.blockSize,
       "define syntax server",
       "q once", &data.once,
       "define syntax client",
       "q count $uint default 1000", &data.count,
       "q node $string", &data.serverNode,
       NULL) != VCLI_OK) {
    vcliPerror (cli, NULL);
    RRR ();
  }
}

/*
 *********************************************************************
 */

int main (int argc, char *argv[]) {

  parseCommandLine (argc, argv);
  if (data.isServer)
    programName = "Socket I/O Test Server";
  else
    programName = "Socket I/O Test Client";

  tsPrint (programName);

  if ((data.vsr = vsocketInit ()) == NULL) RRR ();

  data.buffer = VMEM_REALLOC (data.buffer, data.blockSize);
  if (data.buffer == NULL)
    RRR ();
  else {
    unsigned i;
    for (i = 0; i < data.blockSize; i++) {
      data.buffer[i] = i + 1;
    }
  }

  if (data.isServer)
    server ();
  else
    client ();

  return EXIT_SUCCESS;
}
