/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2000-2004 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/test.c
*
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <vsys_version.h>
#include <vsys_routines.h>
#include <vsys_time.h>
#include <vsys_cli.h>
#include "io.h"
#include "version.h"

static char versionString[] = VSYS_VERSION_STRING "." VERSION_STRING;
static char commandName[] = "vhytec_test";

static unsigned querySize;
static unsigned replySize;
static Vmbport* vp;
static VUINT8* query;
static VUINT8* reply;

enum { CMD_000, CMD_EXEC, CMD_SHOW };
enum { SHOW_000, SHOW_DEVICE, SHOW_FUNCTIONS };

#define SUCCESS (0)
#define FAILURE (-1)

/* command line parameters and qualifiers */

static struct {
  unsigned address;
  unsigned baudRate;
  unsigned debug;
  unsigned numberOfPoints;
  unsigned numTypes;
  unsigned repeatCount;
  unsigned repeatDelay;
  unsigned stop;	/* Number of stop bits. */
  unsigned timeout;
  unsigned unit;
  unsigned value;
  unsigned* valueList;
  char* device;		/* com1, hostname, etc. */
  int* typeList;
  int command;		/* Get or put. */
  int discOrReg;	/* Discrete or register. */
  int enron;
  int inOrOut;		/* Input or output. */
  int hytecFunction;	/* Modbus function number. */
  int parity;		/* VMB_PARITY_*: NONE, EVEN, ODD */
  int port;
  int showWhat;
  int type;		/* VMB_TYPE_*: ASCII, RTU, TCP, ... */
} q;

/* function names according to page 4 of
   http://www.modicon.com/techpubs/dcon7.html */

static struct {
  unsigned code;
  char *name;
} functionTable[] = {
  { 1, "Read Coil Status" },
  { 2, "Read Input Status" },
  { 3, "Read Holding Register" },
  { 4, "Read Input Register" },
  { 5, "Force Single Coil" },
  { 6, "Preset Single Register" },
  { 7, "Read Exception Status" },
#if 0
  { 8, "Diagnostics" },
  { 9, "Program 484" },
  { 10, "Poll 484" },
  { 11, "Fetch Comm. Event Counter" },
  { 12, "Fetch Comm. Event Log" },
  { 13, "Program Controller" },
  { 14, "Poll Controller" },
#endif
  { 15, "Force Multiple Coils" },
  { 16, "Preset Multiple Registers" },
#if 0
  { 17, "Report Slave ID" },
  { 18, "Program 884/M84" },
  { 19, "Reset Comm. Link" },
  { 20, "Read General Reference" },
  { 21, "Write General Reference" },
  { 22, "Mask Write 4x Register" },
  { 23, "Read/Write 4x Register" },
  { 24, "Read FIFO Queue" },
#endif
  { 0, NULL }
};

#define ARRAY_SIZE(array) (sizeof (array) / sizeof (array)[0])

/*
 *********************************************************************
 * Error handling
 *********************************************************************
 */

static int rrr (char *format, ...) {
  vtime bintim;
  char asctim[30];
  va_list args;
  va_start (args, format);
  (void) vsys_get_time (&bintim);
  (void) vsysc_asctime (asctim, sizeof asctim, &bintim);
  (void) fprintf (stderr, "%s: ", asctim);
  (void) vfprintf (stderr, format, args);
  va_end (args);
  (void) fprintf (stderr, "\n");
  (void) fflush (stderr);
  return FAILURE;
}

#define RRR() interr (__LINE__)

static void interr (int line) {
  (void) rrr ("internal error %d", line);
  exit (EXIT_FAILURE);
}

static void errNomem (void) {
  (void) rrr ("insufficient virtual memory");
  exit (EXIT_FAILURE);
}

static int errVmb (void) {
  return rrr ("%s", vmbStrerror (vp));
}

static int errByteCount (unsigned expected, unsigned actual) {
  return rrr ("invalid byte count, expected = %d, actual = %d\n",
	      expected, actual);
}

/*
 *********************************************************************
 */

static int printException (unsigned code) {
  static char *table[] = {
    /* 00 */ NULL,
    /* 01 */ "Illegal Function",
    /* 02 */ "Illegal Data Address",
    /* 03 */ "Illegal Data Value",
    /* #4 is defined differently. In the TCP specification, it is
       "illegal response length". In the other specification,
       it is "slave device failure". */
    /* 04 */ "Illegal Response Length (or Slave Device Failure)",
    /* 05 */ "Acknowledge",
    /* 06 */ "Slave Device Busy",
    /* 07 */ "Negative Acknowledge",
    /* 08 */ "Memory Parity Error",
    /* 09 */ NULL,
    /* 0A */ "Gateway Path Unavailable",
    /* 0B */ "Gateway Target Device Failed To Respond"
  };
  if (code >= ARRAY_SIZE (table) ||
      table[code] == NULL) {
    return rrr ("unknown exception code %02x", code);
  }
  (void) rrr ("%02X - %s", code, table[code]);
  return SUCCESS;
}

/*
 *********************************************************************
 * Common code for all functions
 *********************************************************************
 */

static int fx (unsigned queryLen, unsigned expectedReplyLen) {
  int hldr;
  query[0] = q.hytecFunction;

  if (vmbTransact (vp, q.unit, queryLen, expectedReplyLen, &hldr) != VMB_OK)
    return errVmb ();

  if (reply[0] != q.hytecFunction) {
    if (reply[0] != (q.hytecFunction | 0x80)) {
      RRR ();
      return FAILURE;
    }
    (void) printException (reply[1]);
    return FAILURE;
  }

  return SUCCESS;
}

/*
 *********************************************************************
 * f1 (Read Coil Status) & f2 (Read Input Status)
 *********************************************************************
 */

static int f1f2 (void) {

  unsigned byteCount = (q.numberOfPoints + 7) / 8;
  unsigned expectedReplyLen = byteCount + 2;
  unsigned queryLen = 5;
  unsigned i;

  query[1] = (q.address - 1) >> 8;
  query[2] = (q.address - 1) & 0xff;
  query[3] = q.numberOfPoints >> 8;
  query[4] = q.numberOfPoints & 0xff;

  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  if (reply[1] != byteCount) return errByteCount (byteCount, reply[1]);

  for (i = 0; i < q.numberOfPoints; i++) {
    unsigned byteIndex = i / 8;
    unsigned bitIndex = i % 8;
    if ((i % 16) == 0) {
      if (i != 0) (void) printf ("\n");
      (void) printf ("%05d", q.address + i);
    }
    (void) printf (" %d",
		   (reply[2 + byteIndex] & (1 << bitIndex)) == 0 ? 0 : 1);
  }
  (void) printf ("\n");

  return SUCCESS;
}

/*
 *********************************************************************
 * f3 (Read Input Register) & f4 (Read Holding Register)
 *********************************************************************
 */

static int f3f4 (void) {

  int prevStatus = SUCCESS;
  VUINT8 prev[256 * 2];
  unsigned addr1 = q.address - 1;
  unsigned queryLen = 5;
  unsigned expectedReplyLen;
  unsigned numRegisters = q.numberOfPoints;
  unsigned pass;
  unsigned i;
  int enron = q.enron && addr1 >= 7000 && addr1 <= 7998;

  if (enron) {
    addr1++;
    numRegisters *= 2;
  }

  expectedReplyLen = numRegisters * 2 + 2;

  for (pass = 0; pass < q.repeatCount; pass++) {
    query[1] = addr1 >> 8;
    query[2] = addr1 & 0xff;
    query[3] = q.numberOfPoints >> 8;
    query[4] = q.numberOfPoints & 0xff;

    if (pass != 0 && q.repeatDelay != 0)
      vsys_wait (q.repeatDelay);

    if (fx (queryLen, expectedReplyLen)) {
      if (q.repeatCount == 1)
	return FAILURE;
      else {
	prevStatus = FAILURE;
	continue;
      }
    }

    if (reply[1] != numRegisters * 2)
      return errByteCount (numRegisters * 2, reply[1]);

    if (pass == 0) {
      for (i = 0; i < numRegisters; i++) {
	if ((i % 8) == 0) {
	  if (i != 0) (void) printf ("\n");
	  (void) printf ("%05d", q.address + i);
	}
	(void) printf (" %04X",
		       (reply[2 + i * 2] << 8) | reply[2 + i * 2 + 1]);
      }
      (void) printf ("\n");
      if (q.repeatCount > 1)
	(void) memcpy (&prev, reply + 2, numRegisters * 2);
    }
    else {
      if (prevStatus == FAILURE) {
	prevStatus = SUCCESS;
	(void) rrr ("Reconnected");
      }
      for (i = 0; i < numRegisters; i++) {
	if (reply[2 + i * 2] != prev[i * 2] ||
	    reply[2 + i * 2 + 1] != prev[i * 2 + 1]) {
	  (void) printf ("%05d %04X (pass %d)\n",
			 q.address + i,
			 (reply[2 + i * 2] << 8) | reply[2 + i * 2 + 1],
			 pass);
	  prev[i * 2] = reply[2 + i * 2];
	  prev[i * 2 + 1] = reply[2 + i * 2 + 1];
	}
      }
    }
  }

  if (q.repeatCount != 1)
    (void) printf ("Done %d times\n", q.repeatCount);
  return SUCCESS;
}

/*
 *********************************************************************
 * f5 (Force Single Coil)
 *********************************************************************
 */

static int f5 (void) {

  unsigned queryLen = 5;
  unsigned expectedReplyLen = queryLen;

  query[1] = (q.address - 1) >> 8;
  query[2] = (q.address - 1) & 0xff;
  query[3] = q.value;
  query[4] = 0;

  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  if (memcmp (query, reply, queryLen) != 0) RRR ();
  return SUCCESS;
}

/*
 *********************************************************************
 * f6 (Preset Single Register)
 *********************************************************************
 */

static int f6 (void) {

  unsigned queryLen = 5;
  unsigned expectedReplyLen = queryLen;

  query[1] = (q.address - 1) >> 8;
  query[2] = (q.address - 1) & 0xff;
  query[3] = q.value >> 8;
  query[4] = q.value & 0xff;

  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  if (memcmp (query, reply, queryLen) != 0) RRR ();
  return SUCCESS;
}

/*
 *********************************************************************
 * f7 (Read Exception Status)
 *********************************************************************
 */

static int f7 (void) {
  unsigned queryLen = 1;
  unsigned expectedReplyLen = 2;
  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  (void) printf ("Exception Status = %02X\n", reply[1]);
  return SUCCESS;
}

/*
 *********************************************************************
 * f15 (Force Multiple Coils)
 *********************************************************************
 */

static int f15 (void) {

  unsigned byteCount = (q.numberOfPoints + 7) / 8;
  unsigned queryLen = 6 + byteCount;
  unsigned expectedReplyLen = 5;
  unsigned i;

  query[1] = (q.address - 1) >> 8;
  query[2] = (q.address - 1) & 0xff;
  query[3] = q.numberOfPoints >> 8;
  query[4] = q.numberOfPoints & 0xff;
  query[5] = byteCount;
  if (byteCount >= querySize - 6) RRR ();

  for (i = 0; i < q.numberOfPoints; i++) {
    unsigned byteIndex = i / 8 + 6;
    unsigned bitIndex = i % 8;
    if (bitIndex == 0) query[byteIndex] = 0;
    if (q.valueList[i]) query[byteIndex] |= (1 << bitIndex);
  }

  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  if (memcmp (query, reply, expectedReplyLen) != 0) RRR ();
  return SUCCESS;
}

/*
 *********************************************************************
 * f16 (Preset Multiple Registers)
 *********************************************************************
 */

static int f16 (void) {

  unsigned queryLen = 6 + q.numberOfPoints * 2;
  unsigned expectedReplyLen = 5;
  unsigned i;

  query[1] = (q.address - 1) >> 8;
  query[2] = (q.address - 1) & 0xff;
  query[3] = q.numberOfPoints >> 8;
  query[4] = q.numberOfPoints & 0xff;
  query[5] = q.numberOfPoints * 2;
  if (q.numberOfPoints * 2 >= querySize - 6) RRR ();

  for (i = 0; i < q.numberOfPoints; i++) {
    query[6 + i * 2] = q.valueList[i] >> 8;
    query[6 + i * 2 + 1] = q.valueList[i] & 0xff;
  }

  if (fx (queryLen, expectedReplyLen)) return FAILURE;
  if (memcmp (query, reply, expectedReplyLen) != 0) RRR ();
  return SUCCESS;
}

/*
 *********************************************************************
 */

static int execFunction (void) {

  if ((vp = vmbAlloc (&query, &querySize, &reply, &replySize)) == NULL) {
    errNomem ();
    return FAILURE;
  }
    
  if (vmbSet (vp, VMB_O_DEBUG, (void*) q.debug) != VMB_OK ||
      (q.baudRate && vmbSet (vp, VMB_O_BAUD, (void*) q.baudRate) != VMB_OK) ||
      (q.parity && vmbSet (vp, VMB_O_PARITY, (void*) q.parity) != VMB_OK) ||
      (q.timeout && vmbSet (vp, VMB_O_TIMEOUT, (void*) q.timeout) != VMB_OK) ||
      (q.port && vmbSet (vp, VMB_O_PORT, (void*) q.port) != VMB_OK) ||
      (q.stop && vmbSet (vp, VMB_O_STOP, (void*) q.stop) != VMB_OK) ||
      vmbSet (vp, VMB_O_TYPE, (void*) q.type) != VMB_OK ||
      vmbConnect (vp, q.device) != VMB_OK)
    return errVmb ();
  
  switch (q.hytecFunction) {
  case 1: return f1f2 ();
  case 2: return f1f2 ();
  case 3: return f3f4 ();
  case 4: return f3f4 ();
  case 5: return f5 ();
  case 6: return f6 ();
  case 7: return f7 ();
  case 15: return f15 ();
  case 16: return f16 ();
  default: RRR ();
  }

  return SUCCESS;
}

/*
 *********************************************************************
 */

static int showFunctions (void) {
  unsigned i = 0;
  while (functionTable[i].code != 0) {
    (void) printf ("%2d %s\n", functionTable[i].code, functionTable[i].name);
    i++;
  }
  return SUCCESS;
}

/*
 *********************************************************************
 */

static int showDevice (void) {

  if ((vp = vmbAlloc (&query, &querySize, &reply, &replySize)) == NULL) {
    errNomem ();
    return FAILURE;
  }

  if (vmbSet (vp, VMB_O_DEBUG, (void*) q.debug) != VMB_OK ||
      vmbSet (vp, VMB_O_TYPE, (void*) q.type) != VMB_OK ||
      vmbConnect (vp, q.device) != VMB_OK)
    return errVmb ();

  if (VMB_O_TYPE == VMB_TYPE_TCP) {
    (void) printf ("Nothing for TCP ports\n");
  }
  else {
    int baud, parity, stop;
    char *p;
    if (vmbGet (vp, VMB_O_BAUD, &baud) != VMB_OK ||
	vmbGet (vp, VMB_O_STOP, &stop) != VMB_OK ||
	vmbGet (vp, VMB_O_PARITY, &parity) != VMB_OK)
      return errVmb ();
    (void) printf ("BaudRate = %d\n", baud);
    switch (parity) {
    case SL_PARITY_EVEN: p = "EVEN"; break;
    case SL_PARITY_ODD: p = "ODD"; break;
    case SL_PARITY_NONE: p = "NONE"; break;
    default:  p = "UNKNOWN"; break;
    }
    (void) printf ("Parity   = %s\n", p);
    (void) printf ("StopBits = %d\n", stop);
  }

  return SUCCESS;
}

/*
 *********************************************************************
 */

static void parseArguments (int argc, char **argv) {

  vcli_t *cli;

  if ((cli = vcliInit ()) == NULL) RRR ();

#define DEVICE_QUALIFIERS()						\
  "q baudRate $uint", &q.baudRate,					\
  "q device $string disallow ! device", &q.device,			\
  "q enron", &q.enron,							\
  "q parity type parity", &q.parity,					\
  "q port $uint", &q.port,						\
  "q stop $uint range 1:2", &q.stop,					\
  "q timeout $uint", &q.timeout,					\
  "q type type type list disallow ! type", &q.numTypes, &q.typeList

  if (vcliSetTitle (cli, "Modbus I/O Test Utility") != VCLI_OK ||
      vcliSetVersion (cli, versionString) != VCLI_OK ||
      vcliSetCommandName (cli, commandName) != VCLI_OK ||
      vcliEnableSortedHelp (cli) != VCLI_OK ||
      vcliVparse
      (cli,
       argc, argv,

       "define type command",
       "k execFunction syntax exec", (void*) CMD_EXEC,
       "k show syntax show", (void*) CMD_SHOW,

       "p command optional type command", &q.command,

       "define type type",
       "k ascii", (void*) VMB_TYPE_ASCII,
       "k rtu", (void*) VMB_TYPE_RTU,
       "k tcp", (void*) VMB_TYPE_TCP,

       "define type parity",
       "k NONE", (void*) SL_PARITY_NONE,
       "k EVEN", (void*) SL_PARITY_EVEN,
       "k ODD", (void*) SL_PARITY_ODD,


       "define type hytecFunction",
#define M(n,s) "k f" #n " subsyntax " #s, (void*) n
       M (1, readBinary),
       M (2, readBinary),
       M (3, readInt16),
       M (4, readInt16),
       M (5, writeBinary),
       M (6, writeInt16),
       M (7, f7),
       M (15, f15),
       M (16, f16),
#undef M

       "define syntax exec",
       "p hytecFunction type hytecFunction", &q.hytecFunction,
       "q unit $uint default 1", &q.unit,
       "q debug $uint", &q.debug,
       DEVICE_QUALIFIERS (),

       "define syntax readBinary",
       "p address $uint range 1:65536", &q.address,
       "p numberOrPoints $uint range 1:2040", &q.numberOfPoints,

       "define syntax readInt16",
       "p address $uint range 1:65536", &q.address,
       "p numberOrPoints $uint range 1:125", &q.numberOfPoints,
       "q continuous $uint range 1:2000000000 default 1", &q.repeatCount,
       "q repeatDelay $uint", &q.repeatDelay,
       
       "define syntax writeBinary",
       "p address $uint range 1:65536", &q.address,
       "p value $uint range 0:1", &q.value,

       "define syntax writeInt16",
       "p address $uint range 1:65536", &q.address,
       "p value $uint range 0:65535", &q.value,

       "define syntax f7",

       "define syntax f15",
       "p address $uint range 1:65536", &q.address,
       "p value $uint range 0:1 list", &q.numberOfPoints, &q.valueList,

       "define syntax f16",
       "p address $uint range 1:65536", &q.address,
       "p value $uint range 0:65535 list", &q.numberOfPoints, &q.valueList,

       "define type show",
       "k device syntax showDevice", SHOW_DEVICE,
       "k functions syntax showFunctions", SHOW_FUNCTIONS,

       "define syntax show",
       "p showWhat type show", &q.showWhat,

       "define syntax showDevice",
       "q debug $uint", &q.debug,
       "q device $string disallow ! device", &q.device,
       "q type type type list disallow ! type", &q.numTypes, &q.typeList,

       "define syntax showFunctions",

       NULL) != VCLI_OK) {
    (void) vcliPerror (cli, NULL);
    exit (EXIT_FAILURE);
  }

  if (q.numTypes == 0)
    RRR ();
  else if (q.numTypes == 1)
    q.type = q.typeList[0];
  else if (q.numTypes != 2)
    RRR ();
  else {
    unsigned i0 = q.typeList[0] == VMB_TYPE_TCP ? 0 : 1;
    unsigned i1 = i0 == 0 ? 1 : 0;
    if (q.typeList[i0] != VMB_TYPE_TCP ||
	(q.typeList[i1] != VMB_TYPE_RTU &&
	 q.typeList[i1] != VMB_TYPE_ASCII))
      RRR ();
    q.type = q.typeList[i1] == VMB_TYPE_RTU ?
      VMB_TYPE_TCP_RTU : VMB_TYPE_TCP_ASCII;
  }

#undef DEVICE_QUALIFIERS
}

/*
 *********************************************************************
 */

int main (int argc, char **argv) {

  int status = SUCCESS;

  parseArguments (argc, argv);

  switch (q.command) {
  case CMD_EXEC: status = execFunction (); break;
  case CMD_SHOW:
    switch (q.showWhat) {
    case SHOW_DEVICE: status = showDevice (); break;
    case SHOW_FUNCTIONS: status = showFunctions (); break;
    }
    break;
  }

  return status == SUCCESS ? EXIT_SUCCESS : EXIT_FAILURE;
}
