/******************************************************************************
*        ____
*      _-    -_    _------_
*     -        -_--        -_
*          _--------------_   -_
* 	_---              -_
*      _-                 ----___
*                         ---___
*      Copyright (c) 2001-2003 Vista Control Systems, Inc.
*      All rights reserved
*      176 Central Park Square * Los Alamos, N.M. 87544 * 505-662-2484
*
*      Filename: vhytec/waketest.c
*
**********************************************************************/

/*
 * Program to test/research wake up times.
 * Author: Eugene Zharkov
 * Started: 17-Jul-2001
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <vsys_version.h>
#include <vsys_errors.h>
#include <vsys_time.h>
#include <vsys_cli.h>
#include <vloglib.h>
#include <vthread.h>

#include "version.h"

static char programName[] = "Wakeup Accuracy Test";
static char commandName[] = "waketest";
static char versionString[] = VSYS_VERSION_STRING "." VERSION_STRING;

typedef int bool;

static struct {
  bool banner;
  vtime scanRateTime;		/* Scan rate in vtime form. */
  unsigned scanRate;		/* Scan rate in milliseconds. */
  unsigned maxDiff;		/* Max diff in milliseconds. */
  vthreadCond_t cond;
  vthreadMutex_t mutex;
} data;

/*
 *********************************************************************
 */

#define RRR() rrr (__LINE__)
#define FIXME() rrr (__LINE__)

static void rrr (int line) {
  printf ("vhytec/server: error at %d\n", line);
  exit (EXIT_FAILURE);
}

/*
 *********************************************************************
 */

#define timADD(t1, t2, t3) do {			\
  (t3).t_sec = (t1).t_sec + (t2).t_sec;		\
  (t3).t_nsec = (t1).t_nsec + (t2).t_nsec;	\
  if ((t3).t_nsec >= 1000000000) {		\
    (t3).t_sec += 1;				\
    (t3).t_nsec -= 1000000000;			\
  }						\
} while (0)

/*
 *********************************************************************
 */

static void test (void) {

  struct timespec ts;
  vtime currTime, expTime, prevTime;
  unsigned count = 0;
  unsigned lastReportCount = 0;

  if (vthreadMutexLock (&data.mutex) != VTHREAD_OK) RRR ();

  vsys_get_time (&currTime);

  while (1) {
    count++;
    timADD (currTime, data.scanRateTime, expTime);
    ts.tv_nsec = expTime.t_nsec;
    ts.tv_sec = expTime.t_sec;
    if (vthreadCondTimedwait (&data.cond, &data.mutex, &ts) !=
	VTHREAD_ETIMEDOUT) RRR ();
    vsys_get_time (&currTime);
    if (count > lastReportCount + 1) {
      unsigned diff = (currTime.t_sec - prevTime.t_sec) * 1000;
      if (currTime.t_nsec < prevTime.t_nsec) {
	if (diff == 0) RRR ();
	diff -= 1000;
	diff += (1000000000 + currTime.t_nsec - prevTime.t_nsec) / 1000000;
      }
      else {
	diff += (currTime.t_nsec - prevTime.t_nsec) / 1000000;
      }
      if (diff > data.maxDiff) {
	char a1[40];
	char a2[40];
	lastReportCount++;
	vsysc_asctime (a1, sizeof a1, &prevTime);
	vsysc_asctime (a2, sizeof a2, &currTime);
	printf ("- %s\n  %s\n", a1, a2);
      }
    }
    prevTime = currTime;
  }
}

/*
 *********************************************************************
 */

int main (int argc, char *argv[]) {

  vcli_t * cli;

  if ((cli = vcliInit ()) == NULL) RRR ();

  if (vcliSetTitle (cli, programName) != VCLI_OK ||
      vcliSetVersion (cli, versionString) != VCLI_OK ||
      vcliSetCommandName (cli, commandName) != VCLI_OK ||
      vcliEnableSortedHelp (cli) != VCLI_OK ||
      vcliVparse
      (cli,
       argc, argv,
       "q banner default 1", &data.banner,
       "q rate $uint default 10", &data.scanRate,
       "q maxDiff $uint default 100", &data.maxDiff,
       NULL) != VCLI_OK) {
    vcliPerror (cli, NULL);
    RRR ();
  }

  if (data.banner)
    printf ("Vsystem %s Version %s\n", programName, versionString);

  data.scanRateTime.t_sec = data.scanRate / 1000;
  data.scanRateTime.t_nsec = (data.scanRate % 1000) * 1000000;

  if (vthreadMutexInit (&data.mutex, NULL) != VTHREAD_OK) RRR ();
  if (vthreadCondInit (&data.cond, NULL) != VTHREAD_OK) RRR ();

  test ();
  
  return EXIT_SUCCESS;
}
