/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770	*/
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566	*/
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/
/*                                                                              */
/*				IOC9010 PROJECT	--- Hytec IP Card API Functions 				*/
/*							for Windows and Linux Platforms	    				*/
/*                                IP Card Type: 8510                            */
/********************************************************************************/
/*  Source file name      :-      Hy8510.h	 									*/
/*  Definition header file:		  HYTECIPAPI.h									*/
/*  Platform			  :		  Linux, Windows XP, Windows 2000               */
/*  Initial creation date :-      28-Dec-2011                                   */
/*                                                                              */
/*  Original Developers   :-      Mark Woodward                                 */
/********************************************************************************/
/*                                                                              */
/*  Description :-                                                              */
/*                                                                              */
/*  This header file is designed for Hytec API function library of 8510 digital */
/*	input IP card. It defines the class prototype, error code, masks and 		*/
/*	registers of 8510.														*/
/*																				*/
/*  The design of this interface is cross platforms, which means the code 		*/
/*	can be used in operating systems like Windows, Linux etc without changes.	*/
/*                                                                              */
/*  The language used is C++ but the final product -- the Dynamic Link library	*/
/*	(dll) or the shared library can be used by either C or C++ applications.	*/
/*																				*/
/*  (C)2011 Hytec Electronics Ltd.												*/
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Modification record :-                                                      */
/*                                                                              */
/*  Date modified |     Person    |   Description of modification | Version     */
/* ---------------|---------------|-------------------------------------------- */
/*    28/12/11    |	Mark Woodward |	  Initial copy				  |	0.0.1		*/
/*                |               |                                             */
/*                                                                              */
/********************************************************************************/
#include "HytecIPAPI.h"

#if PLATFORM == 1

#define HYTECAPI_EXPORTS 1
#ifdef HYTECAPI_EXPORTS
#define HYTECAPI_API __declspec(dllexport)
#else
#define HYTECAPI_API __declspec(dllimport)
#endif
#include "IOCGeneric.h"

#else

#include "IOCGeneric.h"

#endif

//error code definition
#define NOT_8510_IP -8510010
#define Hy8510_NOT_CONFIGURED -8510011
#define Hy8510_SCANRATE_INVALID -8510012
#define Hy8510_WATCHDOGRATE_INVALID -8510013
#define Hy8510_MODEMEM_INVALID -8510014
#define Hy8510_MODULESIZE_INVALID -8510015

//signature definition
#define SIGNATURE_8510 0x8510

//register address defines
#define CSR_8510 0x0000
#define INTERRUPT_VECTOR_8510 0x0002
#define SCAN_RATE_8510 0x0004
#define WATCHDOG_RATE_8510 0x0006
#define TIMESTAMP_LSW_8510 0x0008
#define TIMESTAMP_MSW_8510 0x000A
#define ALARM_EVENTS_MEM_8510 0x000C
#define ALARM_EVENTS_READ_8510 0x000E
#define FAULTS_8510 0x0010
#define WD_8510 0x0014

//memory address defines
#define ALARM_EVENT_MEMORY_8510 0x0000
#define ALARM_MODE_MEMORY_8510 0x02000
#define CONTACT_STATUS_MEMORY_8510 0x02080


//bit definitions
#define RESET_8510 0x0001
#define IE_8510 0x0002
#define ARM_8510 0x0004
#define STRIG_8510 0x0008
#define RS_8510 0x0020
#define ENCLR_8510 0x0080
#define EVCLR_8510 0x0100
#define ENABLE_WD_8510 0x8000

#define EVENT_ROW_8510(s)	(s & 0x001F)	
#define EVENT_COL_8510(s)	((s & 0x01E0)>>5)	
#define EVENT_VAL_8510(s)	((s & 0x0800)>>11)	
#define EVENT_COD_8510(s)	((s & 0xF000)>>12)	

//#define ENABLE_INTERRUPT 0x4000
//#define CLEAR_INTERRUPT 0x7FFF


// This class is exported from the HYTECAPI.dll
#if PLATFORM == 1									//windows
class HYTECAPI_API CHy8510 : public CIOCGeneric
#else											//linux
class CHy8510 : public CIOCGeneric
#endif
{
private:
	long m_ErrorCode;							// global error code holder
	int initIPSlot;								// remember the IP slot when configures
	int initVMESlot;
	char m_vector;
	int m_INTLevel;
	int moduleSize;
	USHORT modeData[40];

public:
	CHy8510(void);
	CHy8510(int vmeslot, char ipslot);
	~CHy8510(void);

	long GetErrorCode()
	{
		return m_ErrorCode;
	}
	
/*----------------------------------------------------------------------------------*/
/*Function:		Configure 8510 IP card												*/
/*----------------------------------------------------------------------------------*/
/*Description:	8510 IP configuration routine										*/
/*				according to the settings.											*/
/*Prototype:	long Hy8510_Configure(int vmeslot, char ipslot, int vectornum,		*/
/*				int scanRate, int watchdogRate)										*/
/*Arguements:	vmeslot = VME slot number											*/
/*				ipslot = IP card slot number 0 ~ 3 for VME, 0 ~ 5 for IOC			*/
/*				vectornum = interrupt vector number									*/
/*				size - 0=256 , 1=320												*/
/*				scanRate = 1 ~ 65. see manual										*/
/*				watchdogRate = 1 ~ 65. see manual									*/
/*				ModeData - pointer to 256 or 320 char array							*/
/*				INTLevel = interrupt level. must be 1 ~ 7							*/
/*				callback = user callback function									*/
/*				drive = user data structure passed to the ISR						*/
/*Return:		status code, 0 = successful. negative = errors						*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Configure(int vmeslot, char ipslot, USHORT vectornum, 
						int size, int scanRate, 
						int watchdogRate, char *modeData,
						int INTLevel, void (*callback)(void *), void *drive);

/*----------------------------------------------------------------------------------*/
/*Function:		ARM 8510															*/
/*----------------------------------------------------------------------------------*/
/*Description:	Write to the ARM bit on 8510										*/
/*Prototype:	long Hy8510_ARM(void)												*/
/*Arguements:	value - 1 or 0 to write to ARM bit (only lsb is used)				*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ARM(USHORT value);

/*----------------------------------------------------------------------------------*/
/*Function:		Software Trigger 8510												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Write to the software start trigger bit on 8510						*/
/*Prototype:	long Hy8510_SoftwareTrigger(void)									*/
/*Arguements:	none																*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_SoftwareTrigger(void);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 CSR Register												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 CSR register											*/
/*Prototype:	long Hy8510_ReadCSR(void)											*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadCSR(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 Alarm Event Counts Register (address C)					*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 C register											*/
/*Prototype:	long Hy8510_ReadAlarmEventCount(USHORT *value)						*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadAlarmEventCount(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 Alarm Event Reads Register (address E)					*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 E register											*/
/*Prototype:	long Hy8510_ReadAlarmReadCount(USHORT *value)						*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadAlarmReadCount(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 Fault Register											*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 fault register										*/
/*Prototype:	long Hy8510_ReadFault(void)											*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadFault(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Enable 8510 interrupt												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Enable interrupt for 8510 IP card          							*/
/*Prototype:	long Hy8510_EnableInterrupt(void)									*/
/*Arguements:	none																*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_EnableInterrupt(void);

/*----------------------------------------------------------------------------------*/
/*Function:		Reset 8510 interrupt												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Reset interrupt for 8510 IP card          							*/
/*Prototype:	long Hy8510_ResetInterrupt(void)									*/
/*Arguements:	none																*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ResetInterrupt(void);

/*----------------------------------------------------------------------------------*/
/*Function:		Disable 8510 interrupt             									*/
/*----------------------------------------------------------------------------------*/
/*Description:	Disable interrupt for 8510 IP card          					    */
/*Prototype:	long Hy8510_DisableInterrupt(int void)								*/
/*Arguements:	none																*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_DisableInterrupt(void);

/*----------------------------------------------------------------------------------*/
/*Function:		Read mode data														*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read Current Mode Data Memory of 8510 IP card						*/
/*Prototype:	long Hy8510_Read_Mode_Memory(USHORT *value)							*/
/*Arguements:	*value = pointer to the array which holds the mode data				*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Read_Mode_Memory(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Update mode data													*/
/*----------------------------------------------------------------------------------*/
/*Description:	Update Current Mode Data Memory of 8510 IP card						*/
/*Prototype:	long Hy8510_Update_Mode_Memory(USHORT *value)						*/
/*Arguements:	size = size of matrix - 256 or 320									*/
/*				*value = pointer to the array which holds the mode data				*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Update_Mode_Memory(int size,USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read Alarm Event Data												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the next Alarm Event data of 8510								*/
/*Prototype:	long Hy8510_Read_Alarm_Event(USHORT *value, USHORT offset)			*/
/*Arguements:	*value = pointer to the array which will receice the data			*/
/*				offset = Alarm event to read										*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Read_Alarm_Event(USHORT *value, USHORT offset);

/*----------------------------------------------------------------------------------*/
/*Function:		Read current state of contacts												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the latest set of contacvt states stored by the 8510 card		*/
/*Prototype:	long Hy8510_Read_Contact_Status(USHORT *value)										*/
/*Arguements:	*value = pointer to the array which will receice the data			*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Read_Contact_Status(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read current time stamp												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the latest timestamp from the 2 timestamp registers			*/
/*Prototype:	long Hy8510_Read_TimeStamp(USHORT *value)							*/
/*Arguements:	*value = pointer to the array which will receice the data			*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Read_TimeStamp(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Reset time stamp													*/
/*----------------------------------------------------------------------------------*/
/*Description:	This will clear the 2 timestamp registers							*/
/*Prototype:	long Hy8510_Reset_TimeStamp(USHORT *value)							*/
/*Arguements:	None																*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_Reset_TimeStamp(void);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 Interrupt Vector Register									*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 CSR register											*/
/*Prototype:	long Hy8510_ReadIV(USHORT *value)									*/
/*Arguements:	*value - pointer to location to receive data						*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadIV(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Read 8510 WD Register												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Read the 8510 WD register											*/
/*Prototype:	long Hy8510_ReadWD(void)											*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_ReadWD(USHORT *value);

/*----------------------------------------------------------------------------------*/
/*Function:		Write 8510 WD Register												*/
/*----------------------------------------------------------------------------------*/
/*Description:	Write the 8510 WD register											*/
/*Prototype:	long Hy8510_SetWD(void)											*/
/*Arguements:	*value = pointer to the location to receive data					*/
/*Return:		0 is successfull, negative please see error define					*/
/*----------------------------------------------------------------------------------*/
	long Hy8510_SetWD(USHORT value);


};

#if PLATFORM == 1									//windows
extern "C" HYTECAPI_API CHy8510* CreateCHy8510();
extern "C" HYTECAPI_API CHy8510* CreateCHy8510A(int vmeslot, char ipslot);
extern "C" HYTECAPI_API void DeleteCHy8510( CHy8510* instance );
#else											//linux
extern "C" CHy8510* CreateCHy8510();
extern "C" CHy8510* CreateCHy8510A(int vmeslot, char ipslot);
extern "C" void DeleteCHy8510( CHy8510* instance );
#endif

